import contextlib
import re
import json

from tala.testing.interactions.named_test import InteractionTest
from tala.testing.interactions.turn import UserPassivityTurn, UserInterpretationTurn, SystemUtteranceTurn, \
    RecognitionHypothesesTurn, NotifyStartedTurn, SystemMovesTurn, UserSemanticInputTurn
from tala.model.event_notification import EventNotification


class ParseException(Exception):
    pass


class InteractionTestCompiler(object):
    VALID_TURN_TYPES = ["U>", "S>", "Event>"]

    _turn_matcher = re.compile('^(%s) ?(.*)$' % "|".join(VALID_TURN_TYPES), re.MULTILINE | re.DOTALL)
    _moves_matcher = re.compile(r'([\[{].*[\]}])$')
    _utterance_matcher = re.compile(r'^([\'"]?)(?P<utterance>.*)\1$')
    _confidence_matcher = re.compile(r'^.*\s+(?P<confidence>\$\w+|\d*\.\d+)$')
    _hypothesis_split_re = re.compile(r'\s*\|\s*')

    def compile_interaction_tests(self, filename, file_):
        self._filename = filename
        self._file = file_
        self._result = []
        self._line_number = 0
        self._test = None
        self._eof = False
        while not self._eof:
            line = self._read_multiline()
            if self._eof:
                self._add_current_test()
            else:
                self._process_line(line)
        return self._result

    def _read_multiline(self):
        multiline = ""
        while True:
            line = self._file.readline()
            if line == "":
                self._eof = True
                break
            self._line_number += 1
            if self._is_comment(line):
                continue
            line = line.rstrip(" \r\n")
            if line.endswith("\\"):
                line = line.rstrip("\\")
                multiline += line
                multiline += "\n"
                continue
            multiline += line
            if not line.endswith(";"):
                break
        return multiline

    def _process_line(self, line):
        if self._is_start_of_test(line):
            self._add_current_test()
            testname = self._get_testname(line)
            self._test = InteractionTest(self._filename, testname, [])
        elif self._is_turn(line):
            turn = self._parse_turn(line)
            self._test.turns.append(turn)
        elif line == "":
            pass
        else:
            raise ParseException(
                "Expected content in line %d of '%s' but got '%s'." %
                (self._line_number, self._filename, line.rstrip("\n"))
            )

    def _is_comment(self, line):
        return line.startswith("#")

    def _is_start_of_test(self, line):
        return line.startswith("--- ")

    @staticmethod
    def _is_turn(line):
        m = InteractionTestCompiler._turn_matcher.search(line)
        return m is not None

    def _add_current_test(self):
        if self._test:
            self._result.append(self._test)

    def _get_testname(self, line):
        return line[4:].rstrip("\n")

    def _parse_turn(self, line):
        match = self._turn_matcher.search(line)
        if match:
            turn_type_string, turn_content_as_string = match.groups()
            turn = self._parse_turn_by_type_string(turn_type_string, turn_content_as_string)
            return turn
        else:
            raise ParseException(
                "Expected one of %s on line %d in '%s' but got '%s'." %
                (self.VALID_TURN_TYPES, self._line_number, self._filename, line.rstrip("\n"))
            )

    def _parse_turn_by_type_string(self, type_string, turn_content_as_string):
        if type_string == "U>":
            return self._parse_user_input_turn(turn_content_as_string)
        elif type_string == "S>":
            return self._parse_system_output_turn(turn_content_as_string)
        elif type_string == "Event>":
            return self._parse_event_turn(turn_content_as_string)
        else:
            raise ParseException(
                "Expected one of %s on line %d in '%s' but got '%s'." %
                (self.VALID_TURN_TYPES, self._line_number, self._filename, turn_content_as_string)
            )

    def _parse_user_input_turn(self, turn_content_as_string):
        def is_semantic_input():
            with self._json_guard(turn_content_as_string):
                content = json.loads(turn_content_as_string)
            if isinstance(content, dict):
                return "interpretations" in turn_content_as_string
            return False

        if not turn_content_as_string:
            return UserPassivityTurn(self._line_number)
        if self._is_moves_content_string(turn_content_as_string):
            if is_semantic_input():
                return UserSemanticInputTurn(json.loads(turn_content_as_string), self._line_number)
            moves, utterance, modality = self._parse_moves_content_string(turn_content_as_string)
            return UserInterpretationTurn(moves, self._line_number, utterance=utterance, modality=modality)
        hypothesis_list = self._get_hypothesis_list(turn_content_as_string)
        return RecognitionHypothesesTurn(hypothesis_list, self._line_number)

    def _parse_system_output_turn(self, turn_content_as_string):
        if self._is_moves_content_string(turn_content_as_string):
            moves, utterance, modality = self._parse_moves_content_string(turn_content_as_string)
            return SystemMovesTurn(moves, self._line_number)
        return SystemUtteranceTurn(turn_content_as_string, self._line_number)

    def _get_hypothesis_list(self, string):
        strings = self._hypothesis_split_re.split(string)
        return list(map(self._parse_hypothesis_as_string, strings))

    def _is_moves_content_string(self, string):
        return self._moves_matcher.search(string)

    def _parse_moves_content_string(self, string):
        def create_result(moves, utterance=None, modality=None):
            return moves, utterance, modality

        def extract(json_value):
            if isinstance(json_value, list):
                return create_result(moves=json_value)
            if isinstance(json_value, dict):
                interpretation = json_value
                return create_result(
                    interpretation["moves"],
                    interpretation.get("utterance"),
                    interpretation.get("modality")
                )  # yapf: disable
            raise RuntimeError(f"Expected a list or dict but got {json_value!r}")

        m = self._moves_matcher.search(string)
        if not m:
            raise ParseException(
                "Expected a list of moves or an interpretation object on line %d in '%s' but got '%s'." %
                (self._line_number, self._filename, string)
            )
        (string, ) = m.groups()
        with self._json_guard(string):
            json_value = json.loads(string)
        return extract(json_value)

    @contextlib.contextmanager
    def _json_guard(self, string):
        try:
            yield
        except json.decoder.JSONDecodeError as exception:
            raise ParseException(
                f"Expected valid JSON on line {self._line_number} of '{self._filename}' "
                f"but encountered a decoding error.\n\n"
                f"  Line {self._line_number}: {string}\n\n"
                f"  Error: '{exception}'"
            )

    def _parse_hypothesis_as_string(self, string):
        def confidence():
            match = self._confidence_matcher.search(string)
            return match.group("confidence") if match else None

        def strip_confidence():
            confidence_part = confidence()
            return string.replace(confidence_part, "").strip() if confidence_part else string

        def utterance():
            utterance_part = strip_confidence()
            match = self._utterance_matcher.search(utterance_part)
            return match.group("utterance")

        return utterance(), confidence()

    def _parse_event_turn(self, string):
        try:
            json_dict = json.loads(string)
        except Exception as json_exception:
            raise ParseException(
                "Expected a JSON parseable string on line %d in '%s' but got '%s'. JSON exception: %s" %
                (self._line_number, self._filename, string, json_exception)
            )
        if json_dict["status"] == EventNotification.STARTED:
            return NotifyStartedTurn(json_dict["name"], json_dict["parameters"], self._line_number)
        else:
            raise ParseException(
                f"Expected a supported event turn on line {self._line_number} in '{self._filename}' but got '{string}'."
            )

    @classmethod
    def pretty_passivity(cls):
        return "U>"

    @classmethod
    def pretty_interpretation(cls, interpretation):
        return f"U> {json.dumps(interpretation)}"

    @classmethod
    def pretty_semantic_input(cls, request):
        return f"U> {json.dumps(request)}"

    @classmethod
    def pretty_semantic_expressions(cls, semantic_expressions):
        return f"U> {json.dumps(semantic_expressions)}"

    @classmethod
    def pretty_event(cls, event_dictionary):
        return f"Event> {json.dumps(event_dictionary)}"

    @classmethod
    def pretty_system_utterance(cls, utterance):
        return f"S> {utterance}"

    @classmethod
    def pretty_system_moves(cls, moves):
        return f"S> {json.dumps(moves)}"

    @classmethod
    def pretty_hypotheses(cls, hypotheses):
        def pretty_utterance(utterance):
            return utterance or "''"

        def pretty_hypothesis(utterance, confidence):
            utterance = pretty_utterance(utterance)
            if str(confidence) == "1.0":
                return utterance
            return f"{utterance} {confidence}"

        hypothesis_strings = [pretty_hypothesis(utterance, confidence) for utterance, confidence in hypotheses]
        utterances_with_confidence = " | ".join(hypothesis_strings)
        return f"U> {utterances_with_confidence}"



import unittest

from tala.model.individual import Yes, No


class YesTestCase(unittest.TestCase):
    def test_is_yes(self):
        self.when_created_yes()
        self.then_is_yes_is_true()

    def when_created_yes(self):
        self._yes = Yes()

    def then_is_yes_is_true(self):
        self.assertTrue(self._yes.is_yes())

    def test_hashing_of_yes(self):
        _set = set()
        _set.add(Yes())

    def test_yes_is_positive(self):
        self.assertTrue(Yes().is_positive())


class NoTestCase(unittest.TestCase):
    def test_is_no(self):
        self.when_created_no()
        self.then_is_no_is_true()

    def when_created_no(self):
        self._no = No()

    def then_is_no_is_true(self):
        self.assertTrue(self._no.is_no())

    def test_hashing_of_no(self):
        _set = set()
        _set.add(No())

    def test_no_is_not_positive(self):
        self.assertFalse(No().is_positive())



import json
import unittest
import requests
import logging

from mock import patch, Mock

from tala.ddd.services.service_interface import ServiceParameter
from tala.model.individual import Individual
from tala.model.predicate import Predicate
from tala.model.proposition import PredicateProposition
from tala.model.polarity import Polarity
from tala.model.question import WhQuestion
from tala.model.service_query_result import QueryResultFromService
from tala.model.service_action_outcome import SuccessfulServiceAction, FailedServiceAction
from tala.model.service_invocation import Context, PROTOCOL_VERSION
from tala.model.sort import Sort

import tala.ddd.services.http
from tala.ddd.ddd_specific_components import DDDSpecificComponents
from tala.ddd.services.http import HttpServiceClient, HttpServiceInvocationException
from tala.ddd.services.parameters.binding import SingleInstanceParameterBinding


class HttpServiceClientTest(unittest.TestCase):
    def setUp(self):
        self._requests_patcher = patch(f"{tala.ddd.services.http.__name__}")
        self._mock_logger = Mock(spec=logging.Logger)
        self._mock_session = self._requests_patcher.start()
        self._mock_ddd_components = self._create_mock_ddd_components()
        self._mock_parameter_bindings = [
            SingleInstanceParameterBinding(
                parameter=self._mock_service_parameter("mock_predicate_name"),
                value="mock_value",
                proposition=self.mock_predicate_proposition(
                    predicate_name="mock_predicate_name", sort_name="mock_sort_name", value="mock_value"
                )
            )
        ]

    def tearDown(self):
        self._requests_patcher.stop()

    def _expected_facts_as_dict(
        self,
        sort="mock_sort_name",
        grammar_entry="mock_grammar_entry",
        value="mock_value",
        perception_confidence=None,
        understanding_confidence=None,
        weighted_confidence=None,
        weighted_understanding_confidence=None
    ):
        return {
            "mock_predicate_name": {
                "sort": sort,
                "grammar_entry": grammar_entry,
                "perception_confidence": perception_confidence,
                "understanding_confidence": understanding_confidence,
                "weighted_confidence": weighted_confidence,
                "weighted_understanding_confidence": weighted_understanding_confidence,
                "value": value
            }
        }

    def _expected_parameters_as_dict(
        self,
        sort="mock_sort_name",
        grammar_entry="mock_grammar_entry",
        value="mock_value",
        perception_confidence=None,
        understanding_confidence=None,
        weighted_confidence=None,
        weighted_understanding_confidence=None
    ):
        return {
            "mock_predicate_name": {
                "sort": sort,
                "grammar_entry": grammar_entry,
                "perception_confidence": perception_confidence,
                "understanding_confidence": understanding_confidence,
                "weighted_confidence": weighted_confidence,
                "weighted_understanding_confidence": weighted_understanding_confidence,
                "value": value
            }
        }

    def _create_mock_ddd_components(self):
        mock_components = Mock(spec=DDDSpecificComponents)
        return mock_components

    def _mock_service_parameter(self, name):
        mock_service_parameter = Mock(spec=ServiceParameter)
        mock_service_parameter.name = name
        return mock_service_parameter

    def given_http_service_client(self, endpoint, name):
        self._http_service_client = HttpServiceClient(self._mock_logger, endpoint, name)
        self._http_service_client._session = self._mock_session

    def then_request_is_posted(self, expected_url, expected_data, expected_headers):
        self._mock_session.post.assert_called_once_with(
            expected_url, data=expected_data, headers=expected_headers, timeout=2
        )

    def given_requests_post_returns_response(self, response_text):
        self._mock_session.post.return_value = self._mock_response(response_text)

    def _mock_response(self, text):
        mock_response = Mock(spec=requests.Response)
        mock_response.text = text
        return mock_response

    def then_result_is(self, expected):
        self.assertEqual(expected, self._actual_result)

    def test_perform_posts_request(self):
        self.given_http_service_client("mock_endpoint", "mock_action")
        self.given_requests_post_returns_response(
            json.dumps({
                "status": "success",
                "data": {
                    "version": PROTOCOL_VERSION,
                }
            })
        )
        self.given_session([{
            "sort": "mock_sort",
            "natural_language_form": "mock_grammar_entry",
            "name": "mock_value"
        }], "mock_session_id")

        self.when_perform(
            "mock_action", self._mock_parameter_bindings,
            Context("mock_active_ddd", self._mock_facts(), "mock_invocation_id")
        )
        self.then_request_is_posted(
            expected_url="mock_endpoint",
            expected_data=json.dumps({
                "version": PROTOCOL_VERSION,
                "session": {
                    "session_id": "mock_session_id",
                    "entities": [{
                        "sort": "mock_sort",
                        "natural_language_form": "mock_grammar_entry",
                        "name": "mock_value"
                    }]
                },
                "request": {
                    "type": "request",
                    "name": "mock_action",
                    "parameters": self._expected_parameters_as_dict()
                },
                "context": {
                    "active_ddd": "mock_active_ddd",
                    "facts": self._expected_facts_as_dict(),
                    "invocation_id": "mock_invocation_id"
                }
            }),
            expected_headers={"Content-type": "application/json"}
        )

    def given_session(self, entities, session_id):
        self._session = {"session_id": session_id, "entities": entities}

    def when_perform(self, action, parameters, context):
        self._actual_result = self._http_service_client.perform(action, parameters, self._session, context)

    def test_perform_returns_successful_result(self):
        self.given_http_service_client("mock_endpoint", "mock_action")
        self.given_requests_post_returns_response(
            json.dumps({
                "status": "success",
                "data": {
                    "version": PROTOCOL_VERSION,
                }
            })
        )
        self.given_session([], "mock_session_id")
        self.when_perform(
            "mock_action", self._mock_parameter_bindings,
            Context("mock_active_ddd", self._mock_facts(), "mock_invocation_id")
        )
        self.then_result_is(SuccessfulServiceAction())

    def test_perform_returns_failed_result(self):
        self.given_http_service_client("mock_endpoint", "mock_action")
        self.given_requests_post_returns_response(
            json.dumps({
                "status": "fail",
                "data": {
                    "version": PROTOCOL_VERSION,
                    "reason": "mock_failure_reason"
                }
            })
        )
        self.given_session([], "mock_session_id")
        self.when_perform(
            "mock_action", self._mock_parameter_bindings,
            Context("mock_active_ddd", self._mock_facts(), "mock_invocation_id")
        )
        self.then_result_is(FailedServiceAction("mock_failure_reason"))

    def test_query_posts_request(self):
        self.given_http_service_client("mock_endpoint", "mock_question_predicate")
        self.given_requests_post_returns_response(
            json.dumps({
                "status": "success",
                "data": {
                    "version": PROTOCOL_VERSION,
                    "result": []
                }
            })
        )
        self.given_session([{
            "sort": "mock_sort",
            "natural_language_form": "mock_grammar_entry",
            "name": "mock_value"
        }], "mock_session_id")

        self.when_query(
            question=self.mock_wh_question("mock_question_predicate"),
            parameters=self._mock_parameter_bindings,
            min_results=0,
            max_results=None,
            session=self._session,
            context=Context("mock_active_ddd", self._mock_facts(), "mock_invocation_id")
        )
        self.then_request_is_posted(
            expected_url="mock_endpoint",
            expected_data=json.dumps({
                "version": PROTOCOL_VERSION,
                "session": {
                    "session_id": "mock_session_id",
                    "entities": [{
                        "sort": "mock_sort",
                        "natural_language_form": "mock_grammar_entry",
                        "name": "mock_value"
                    }]
                },
                "request": {
                    "type": "query",
                    "name": "mock_question_predicate",
                    "parameters": self._expected_parameters_as_dict(),
                    "min_results": 0,
                    "max_results": None
                },
                "context": {
                    "active_ddd": "mock_active_ddd",
                    "facts": self._expected_facts_as_dict(),
                    "invocation_id": "mock_invocation_id"
                }
            }),
            expected_headers={"Content-type": "application/json"}
        )

    def mock_wh_question(self, predicate_name, sort_name="mock_sort"):
        mock_question = Mock(spec=WhQuestion)
        sort = self._mock_sort(sort_name)
        predicate = self._mock_predicate(predicate_name, sort)
        mock_question.predicate = predicate
        return mock_question

    def when_query(self, *args, **kwargs):
        self._actual_result = self._http_service_client.query(*args, **kwargs)

    def test_query_returns_parsed_response(self):
        self.given_http_service_client("mock_endpoint", "mock_question_predicate")
        self.given_requests_post_returns_response(
            json.dumps({
                "status": "success",
                "data": {
                    "version": PROTOCOL_VERSION,
                    "result": [{
                        "value": "mock_value",
                        "confidence": "mock_confidence",
                        "grammar_entry": "mock_grammar_entry"
                    }]
                }
            })
        )
        self.when_query(
            question=self.mock_wh_question("mock_question_predicate"),
            parameters=self._mock_parameter_bindings,
            min_results=0,
            max_results=None,
            session={'session_id': 'mock_session_id'},
            context=Context("mock_active_ddd", self._mock_facts(), "mock_invocation_id")
        )
        self.then_result_is([QueryResultFromService("mock_value", "mock_confidence", "mock_grammar_entry")])

    def test_request_for_validate_with_grammar_entry(self):
        self.given_http_service_client("mock_endpoint", "MockValidator")
        self.given_requests_post_returns_response(
            json.dumps({
                "status": "success",
                "data": {
                    "version": PROTOCOL_VERSION,
                    "is_valid": True,
                }
            })
        )
        self.given_session([{
            "sort": "mock_sort",
            "natural_language_form": "mock_grammar_entry",
            "name": "mock_value"
        }], "mock_session_id")
        self.when_validate(
            "MockValidator", self._mock_parameter_bindings,
            Context("mock_active_ddd", self._mock_facts(), "mock_invocation_id")
        )
        self.then_request_is_posted(
            expected_url="mock_endpoint",
            expected_data=json.dumps({
                "version": PROTOCOL_VERSION,
                "session": {
                    "session_id": "mock_session_id",
                    "entities": [{
                        "sort": "mock_sort",
                        "natural_language_form": "mock_grammar_entry",
                        "name": "mock_value"
                    }]
                },
                "request": {
                    "type": "validator",
                    "name": "MockValidator",
                    "parameters": self._expected_parameters_as_dict()
                },
                "context": {
                    "active_ddd": "mock_active_ddd",
                    "facts": self._expected_facts_as_dict(),
                    "invocation_id": "mock_invocation_id"
                }
            }),
            expected_headers={"Content-type": "application/json"}
        )

    def when_validate(self, validator_name, parameters, context):
        self._actual_result = self._http_service_client.validate(validator_name, parameters, self._session, context)

    def mock_predicate_proposition(
        self,
        predicate_name,
        sort_name,
        value,
        polarity=Polarity.POS,
        perception_confidence=None,
        understanding_confidence=None,
        weighted_confidence=None,
        weighted_understanding_confidence=None
    ):
        sort = self._mock_sort(sort_name)
        predicate = self._mock_predicate(predicate_name, sort)
        individual = self._mock_individual(value, sort)
        mock_predicate_proposition = Mock(spec=PredicateProposition)
        mock_predicate_proposition.is_predicate_proposition.return_value = True
        mock_predicate_proposition.get_polarity.return_value = polarity
        mock_predicate_proposition.predicate = predicate
        mock_predicate_proposition.individual = individual
        mock_predicate_proposition.confidence_estimates.perception_confidence = perception_confidence
        mock_predicate_proposition.confidence_estimates.understanding_confidence = understanding_confidence
        mock_predicate_proposition.confidence_estimates.weighted_confidence = weighted_confidence
        mock_predicate_proposition.confidence_estimates.weighted_understanding_confidence = weighted_understanding_confidence
        return mock_predicate_proposition

    def _mock_facts(
        self,
        predicate_name="mock_predicate_name",
        sort_name="mock_sort_name",
        value="mock_value",
        polarity=Polarity.POS,
        perception_confidence=None,
        understanding_confidence=None,
        weighted_confidence=None,
        weighted_understanding_confidence=None
    ):
        return [
            self.mock_predicate_proposition(
                predicate_name=predicate_name,
                sort_name=sort_name,
                value=value,
                polarity=polarity,
                perception_confidence=perception_confidence,
                understanding_confidence=understanding_confidence,
                weighted_confidence=weighted_confidence,
                weighted_understanding_confidence=weighted_understanding_confidence
            )
        ]

    def _mock_predicate(self, name, sort):
        mock_predicate = Mock(spec=Predicate)
        mock_predicate.get_name.return_value = name
        mock_predicate.sort = sort
        return mock_predicate

    def _mock_sort(self, name):
        mock_sort = Mock(spec=Sort)
        mock_sort.get_name.return_value = name
        return mock_sort

    def _mock_individual(self, value, sort):
        mock_individual = Mock(spec=Individual)
        mock_individual.sort = sort
        mock_individual.value = value
        mock_individual.value_as_json_object.return_value = {"value": value}
        return mock_individual

    def test_request_for_validate_without_grammar_entry(self):
        self.given_http_service_client("mock_endpoint", "MockValidator")
        self.given_requests_post_returns_response(
            json.dumps({
                "status": "success",
                "data": {
                    "version": PROTOCOL_VERSION,
                    "is_valid": True,
                }
            })
        )
        self.given_session([], "mock_session_id")
        self.when_validate(
            "MockValidator", self._mock_parameter_bindings,
            Context("mock_active_ddd", self._mock_facts(), "mock_invocation_id")
        )
        self.then_request_is_posted(
            expected_url="mock_endpoint",
            expected_data=json.dumps({
                "version": PROTOCOL_VERSION,
                "session": {
                    "session_id": "mock_session_id",
                    "entities": []
                },
                "request": {
                    "type": "validator",
                    "name": "MockValidator",
                    "parameters": self._expected_parameters_as_dict(grammar_entry=None)
                },
                "context": {
                    "active_ddd": "mock_active_ddd",
                    "facts": self._expected_facts_as_dict(grammar_entry=None),
                    "invocation_id": "mock_invocation_id"
                }
            }),
            expected_headers={"Content-type": "application/json"}
        )

    def test_successful_validation_response(self):
        self.given_http_service_client("mock_endpoint", "MockValidator")
        self.given_requests_post_returns_response(
            json.dumps({
                "status": "success",
                "data": {
                    "version": PROTOCOL_VERSION,
                    "is_valid": True
                }
            })
        )
        self.given_session([], "mock_session_id")
        self.when_validate(
            "MockValidator", self._mock_parameter_bindings,
            Context("mock_active_ddd", self._mock_facts(), "mock_invocation_id")
        )
        self.then_result_is(True)

    def test_failed_validation_response(self):
        self.given_http_service_client("mock_endpoint", "MockValidator")
        self.given_requests_post_returns_response(
            json.dumps({
                "status": "success",
                "data": {
                    "version": PROTOCOL_VERSION,
                    "is_valid": False,
                }
            })
        )
        self.given_session([], "mock_session_id")
        self.when_validate(
            "MockValidator", self._mock_parameter_bindings,
            Context("mock_active_ddd", self._mock_facts(), "mock_invocation_id")
        )
        self.then_result_is(False)

    def test_fail_with_query(self):
        self.given_http_service_client("mock_endpoint", "MockQuery")
        self.given_requests_post_returns_response(
            json.dumps({
                "status": "fail",
                "data": {
                    "version": PROTOCOL_VERSION,
                }
            })
        )
        self.when_query_then_exception_is_raised_matching(
            question=self.mock_wh_question("mock_question_predicate"),
            parameters=self._mock_parameter_bindings,
            min_results=0,
            max_results=None,
            session={'session_id': 'mock_session_id'},
            context=Context("mock_active_ddd", self._mock_facts(), "mock_invocation_id"),
            expected_exception=HttpServiceInvocationException,
            expected_message="Expected status 'success' but got 'fail'."
        )

    def when_query_then_exception_is_raised_matching(self, expected_exception, expected_message, *args, **kwargs):
        with self.assertRaisesRegex(expected_exception, expected_message):
            self._http_service_client.query(*args, **kwargs)

    def test_fail_with_validator(self):
        self.given_http_service_client("mock_endpoint", "MockValidator")
        self.given_requests_post_returns_response(
            json.dumps({
                "status": "fail",
                "data": {
                    "version": PROTOCOL_VERSION,
                }
            })
        )
        self.when_validating_then_exception_is_raised_matching(
            validator_name="MockValidator",
            parameters=self._mock_parameter_bindings,
            session={'session_id': 'mock_session_id'},
            context=Context("mock_active_ddd", self._mock_facts(), "mock_invocation_id"),
            expected_exception=HttpServiceInvocationException,
            expected_message="Expected status 'success' but got 'fail'."
        )

    def when_validating_then_exception_is_raised_matching(self, expected_exception, expected_message, *args, **kwargs):
        with self.assertRaisesRegex(expected_exception, expected_message):
            self._http_service_client.validate(*args, **kwargs)

    def test_error_with_validator(self):
        self.given_http_service_client("mock_endpoint", "MockValidator")
        self.given_requests_post_returns_response(
            json.dumps({
                "status": "error",
                "message": "an error message",
                "data": {
                    "version": PROTOCOL_VERSION,
                }
            })
        )
        self.when_validating_then_exception_is_raised_matching(
            validator_name="MockValidator",
            parameters=self._mock_parameter_bindings,
            session={'session_id': 'mock_session_id'},
            context=Context("mock_active_ddd", self._mock_facts(), "mock_invocation_id"),
            expected_exception=HttpServiceInvocationException,
            expected_message="Expected service to succeed but it had an error with message: 'an error message'."
        )

    def test_error_with_perform(self):
        self.given_http_service_client("mock_endpoint", "mock_action")
        self.given_requests_post_returns_response(
            json.dumps({
                "status": "error",
                "message": "an error message",
                "data": {
                    "version": PROTOCOL_VERSION,
                }
            })
        )
        self.when_perform_then_exception_is_raised_matching(
            action="mock_action",
            parameters=self._mock_parameter_bindings,
            session={'session_id': 'mock_session_id'},
            context=Context("mock_active_ddd", self._mock_facts(), "mock_invocation_id"),
            expected_exception=HttpServiceInvocationException,
            expected_message="Expected service to succeed but it had an error with message: 'an error message'."
        )

    def when_perform_then_exception_is_raised_matching(self, expected_exception, expected_message, *args, **kwargs):
        with self.assertRaisesRegex(expected_exception, expected_message):
            self._http_service_client.perform(*args, **kwargs)

    def test_unexpected_version(self):
        self.given_http_service_client("mock_endpoint", "mock_action")
        self.given_requests_post_returns_response(json.dumps({
            "status": "success",
            "data": {
                "version": "0.0",
            }
        }))
        self.when_perform_then_exception_is_raised_matching(
            action="mock_action",
            parameters=self._mock_parameter_bindings,
            session={'session_id': 'mock_session_id'},
            context=Context("mock_active_ddd", self._mock_facts(), "mock_invocation_id"),
            expected_exception=HttpServiceInvocationException,
            expected_message=r"Expected one of the supported versions \['1.0', '1.1'\] but got '0.0'"
        )

    def test_request_containing_confidence_scores(self):
        self.given_http_service_client("mock_endpoint", "MockValidator")
        self.given_requests_post_returns_response(
            json.dumps({
                "status": "success",
                "data": {
                    "version": PROTOCOL_VERSION,
                    "result": [{
                        "value": "mock_value",
                        "confidence": "mock_confidence",
                        "grammar_entry": "mock_grammar_entry"
                    }]
                }
            })
        )  # yapf: disable
        self.when_query(
            question=self.mock_wh_question("mock_question_predicate"),
            parameters=self._mock_parameter_bindings,
            min_results=0,
            max_results=None,
            session={'session_id': 'mock_session_id'},
            context=Context(
                "mock_active_ddd",
                self._mock_facts(
                    perception_confidence=0.5,
                    understanding_confidence=1.0,
                    weighted_confidence=0.6,
                    weighted_understanding_confidence=1.2
                ), "mock_invocation_id"
            )
        )
        self.then_result_is([QueryResultFromService("mock_value", "mock_confidence", "mock_grammar_entry")])



import copy

from tala.model.ask_feature import AskFeature
from tala.model.domain import Domain
from tala.model.action import Action
from tala.model.goal import ResolveGoal, PerformGoal
from tala.model.lambda_abstraction import LambdaAbstractedPredicateProposition
from tala.model import speaker
from tala.model.ontology import Ontology
from tala.model.plan import Plan, InvalidPlansException
from tala.model.plan_item import Bind, IfThenElse, InvokeServiceAction, GetDone, Findout, Raise
from tala.model.polarity import Polarity
from tala.model.proposition import PredicateProposition, GoalProposition, PropositionSet, ServiceActionTerminatedProposition
from tala.model.question import WhQuestion, AltQuestion, KnowledgePreconditionQuestion
from tala.model.sort import CustomSort, RealSort
from tala.testing.lib_test_case import LibTestCase
from tala.ddd.json_parser import JSONDomainParser


class DomainTests(LibTestCase):
    def setUp(self):
        self.ontology_name = "mockup_ontology"
        self._city_sort = CustomSort(self.ontology_name, "city", dynamic=True)
        self.setUpDomainTestCase()

    def setUpDomainTestCase(self):
        self._create_ontologies()
        self._create_semantic_objects()
        self._create_domains()

    def _create_ontologies(self):
        sorts = {
            CustomSort(self.ontology_name, "country"),
            self._city_sort,
            CustomSort(self.ontology_name, "city_type"),
        }
        predicates = {
            self._create_predicate("dest_city", self._city_sort),
            self._create_predicate("dept_city", self._city_sort),
            self._create_predicate("dest_country", CustomSort(self.ontology_name, "country")),
            self._create_predicate("price", RealSort()),
            self._create_predicate("code", RealSort()),
            self._create_predicate(
                "dest_city_type", sort=CustomSort(self.ontology_name, "city_type"), feature_of_name="dest_city"
            )
        }
        individuals = {
            "paris": self._city_sort,
            "london": self._city_sort,
        }
        self.actions = {
            "top",
            "buy",
            "always_preferred",
            "conditionally_preferred",
            "planless",
            "with_context",
            "accommodate_unrestricted",
            "downdate_plan_true",
            "downdate_plan_false",
            "event",
            "instructional",
            "user_targeted_action",
        }
        self.ontology = Ontology(self.ontology_name, sorts, predicates, individuals, self.actions)

        self.empty_ontology = Ontology("empty_ontology", {}, {}, {}, set())

    def _create_semantic_objects(self):
        self.predicate_dest_city = self.ontology.get_predicate("dest_city")
        self.predicate_dest_city_type = self.ontology.get_predicate("dest_city_type")
        self.predicate_dept_city = self.ontology.get_predicate("dept_city")
        self.predicate_price = self.ontology.get_predicate("price")
        self.predicate_dest_country = self.ontology.get_predicate("dest_country")
        self.predicate_code = self.ontology.get_predicate("code")

        self.lambda_abstracted_price_prop = LambdaAbstractedPredicateProposition(
            self.predicate_price, self.ontology_name
        )
        self.lambda_abstracted_dest_city_prop = LambdaAbstractedPredicateProposition(
            self.predicate_dest_city, self.ontology_name
        )
        self.lambda_abstracted_dest_city_type_prop = LambdaAbstractedPredicateProposition(
            self.predicate_dest_city_type, self.ontology_name
        )
        self.lambda_abstracted_dept_city_prop = LambdaAbstractedPredicateProposition(
            self.predicate_dept_city, self.ontology_name
        )

        self.price_question = WhQuestion(self.lambda_abstracted_price_prop)
        self.dest_city_question = WhQuestion(self.lambda_abstracted_dest_city_prop)
        self.dest_city_type_question = WhQuestion(self.lambda_abstracted_dest_city_type_prop)
        self.dept_city_question = WhQuestion(self.lambda_abstracted_dept_city_prop)
        self.dest_country_question = WhQuestion(
            LambdaAbstractedPredicateProposition(self.predicate_dest_country, self.ontology_name)
        )

        self.individual_paris = self.ontology.create_individual("paris")
        self.individual_london = self.ontology.create_individual("london")
        self.proposition_dest_city_paris = PredicateProposition(self.predicate_dest_city, self.individual_paris)
        self.proposition_dest_city_london = PredicateProposition(self.predicate_dest_city, self.individual_london)
        self.proposition_not_dest_city_paris = PredicateProposition(
            self.predicate_dest_city, self.individual_paris, Polarity.NEG
        )
        self.instructional = self.ontology.create_action("instructional")

    def _create_domains(self):
        self.domain_name = "mockup_domain"

        self._create_actions()
        self._create_questions()
        self.condition = self.proposition_dest_city_paris
        plans = self._create_plans()

        dependencies = {self.dest_country_question: {self.dest_city_question}}

        self.non_verbalized_question = self.dest_city_question

        parameters = {
            self.non_verbalized_question: {
                "verbalize": False
            },
        }

        self.domain = Domain(
            self.DDD_NAME,
            self.domain_name,
            self.ontology,
            plans=plans,
            default_questions=[self.dest_city_question],
            dependencies=dependencies,
            parameters=parameters
        )
        domain_as_json = self.domain.as_json()
        json_parser = JSONDomainParser(self.ontology)
        self.domain = json_parser.parse(domain_as_json)

        self.empty_domain = Domain(self.DDD_NAME, "empty_domain", self.empty_ontology)

    def _create_actions(self):
        for action_name in self.actions:
            action = self.ontology.create_action(action_name)
            attribute_name = action_name + "_action"
            setattr(self, attribute_name, action)
        self.up_action = self.ontology.create_action("up")

    def _create_questions(self):
        for predicate_name in ["dest_city", "dept_city", "code"]:
            question = self.ontology.create_wh_question(predicate_name)
            attribute_name = predicate_name + "_question"
            setattr(self, attribute_name, question)

    def _create_plans(self):
        return [{
            "goal": ResolveGoal(self.price_question, speaker.SYS),
            "plan": Plan([
                Findout(self.domain_name, self.dest_city_question),
                Findout(self.domain_name, self.dept_city_question),
            ]),
        }, {
            "goal": PerformGoal(self.buy_action),
            "plan": Plan([Findout(self.domain_name, self.price_question)]),
            "postconds": [self.condition]
        }, {
            "goal": PerformGoal(self.always_preferred_action),
            "preferred": True,
            "plan": Plan([]),
            "postplan": [Findout(self.domain_name, self.price_question)],
            "superactions": [self.buy_action]
        }, {
            "goal": PerformGoal(self.conditionally_preferred_action),
            "preferred": self.condition,
            "plan": Plan([])
        }, {
            "goal": PerformGoal(self.accommodate_unrestricted_action),
            "plan": Plan([]),
            "unrestricted_accommodation": True,
        }, {
            "goal": PerformGoal(self.downdate_plan_true_action),
            "plan": Plan([InvokeServiceAction("mockup_ontology", "mock_service_action", downdate_plan=True)]),
        }, {
            "goal": PerformGoal(self.downdate_plan_false_action),
            "plan": Plan([InvokeServiceAction("mockup_ontology", "mock_service_action", downdate_plan=False)]),
        }, {
            "goal": PerformGoal(self.instructional_action),
            "plan": Plan([GetDone(Action("user_targeted_action", "mockup_ontology"))]),
        }]

    def test_top_and_up_goals_added_implicitly_by_domain(self):
        self.given_custom_goals_without_top_and_up()
        self.when_fetching_goals_from_domain()
        self.then_top_and_up_are_included()

    def given_custom_goals_without_top_and_up(self):
        plans = self._create_plans()
        goals = [info["goal"] for info in plans]
        self.expected_goals = goals
        action_names = self._action_names_of_perform_goals(self.expected_goals)
        self.assertNotIn("top", action_names)
        self.assertNotIn("up", action_names)

    def when_fetching_goals_from_domain(self):
        self.actual_goals = self.domain.goals

    def then_top_and_up_are_included(self):
        self.assertEqual(len(self.actual_goals), len(self.expected_goals) + 2)
        action_names = self._action_names_of_perform_goals(self.actual_goals)
        self.assertIn("top", action_names)
        self.assertIn("up", action_names)

    def _action_names_of_perform_goals(self, goals):
        actions_of_perform_goals = [goal.action for goal in goals if goal.is_perform_goal()]
        action_names = [action.value for action in actions_of_perform_goals]
        return action_names

    def test_domain_equality(self):
        identical_domain = copy.copy(self.domain)
        self.assert_eq_returns_true_and_ne_returns_false_symmetrically(self.domain, identical_domain)

    def test_domain_inequality(self):
        non_identical_domain = self.empty_domain
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(self.domain, non_identical_domain)

    def test_get_name(self):
        self.assertEqual("mockup_domain", self.domain.get_name())

    def test_action_method(self):
        self.assertEqual(self.buy_action, self.domain.action("buy"))

    def test_has_goal_true_for_question_with_plan(self):
        self.assertTrue(self.domain.has_goal(ResolveGoal(self.price_question, speaker.SYS)))

    def test_has_goal_false_for_question_without_plan(self):
        question_without_plan = self.dest_city_question
        self.assertFalse(self.domain.has_goal(ResolveGoal(question_without_plan, speaker.USR)))

    def test_has_goal_true_for_action_with_plan(self):
        self.assertTrue(self.domain.has_goal(PerformGoal(self.buy_action)))

    def test_has_goal_false_for_planless_action(self):
        self.assertFalse(self.domain.has_goal(PerformGoal(self.planless_action)))

    def test_up_plan_empty(self):
        self.assertEqual(0, len(self.domain.get_plan(PerformGoal(self.up_action))))

    def test_goal_is_preferred_true_if_plan_always_preferred(self):
        self.assertTrue(self.domain.goal_is_preferred(PerformGoal(self.always_preferred_action)))

    def test_goal_is_preferred_false_if_plan_never_preferred(self):
        never_preferred_action = self.buy_action
        self.assertFalse(self.domain.goal_is_preferred(PerformGoal(never_preferred_action)))

    def test_goal_is_preferred_true_if_condition_in_facts(self):
        facts_with_condition = [self.condition]
        self.assertTrue(
            self.domain.goal_is_conditionally_preferred(
                PerformGoal(self.conditionally_preferred_action), facts_with_condition
            )
        )

    def test_goal_is_preferred_false_if_condition_not_in_facts(self):
        facts_without_condition = []
        self.assertFalse(
            self.domain.goal_is_conditionally_preferred(
                PerformGoal(self.conditionally_preferred_action), facts_without_condition
            )
        )

    def test_is_default_question(self):
        default_question = self.dest_city_question
        non_default_question = self.price_question
        self.assertTrue(self.domain.is_default_question(default_question))
        self.assertFalse(self.domain.is_default_question(non_default_question))

    def test_plan_can_be_silently_accommodated(self):
        silently_accommodatable_goal = PerformGoal(self.accommodate_unrestricted_action)
        self.assertTrue(self.domain.is_silently_accommodatable(silently_accommodatable_goal))

    def test_plan_with_conditionally_preferred_action_can_not_be_silently_accommodated(self):
        silently_accommodatable_goal = PerformGoal(self.conditionally_preferred_action)
        self.assertFalse(self.domain.is_silently_accommodatable(silently_accommodatable_goal))

    def test_get_plan_questions(self):
        self.maxDiff = None
        expected_questions = set(["?X.dept_city(X)", "?X.dest_city_type(X)", "?X.dest_city(X)", "?X.price(X)"])
        actual_result = [question for question in self.domain.get_plan_questions()]
        actual_result_strings = set(map(str, actual_result))
        assert expected_questions == actual_result_strings

    def test_get_all_goals(self):
        expected_goals = {
            ResolveGoal(self.price_question, speaker.SYS),
            PerformGoal(self.buy_action),
            PerformGoal(self.always_preferred_action),
            PerformGoal(self.conditionally_preferred_action),
            PerformGoal(self.up_action),
            PerformGoal(self.accommodate_unrestricted_action),
            PerformGoal(self.top_action),
            PerformGoal(self.downdate_plan_true_action),
            PerformGoal(self.downdate_plan_false_action),
            PerformGoal(self.instructional),
        }
        actual_result = set(self.domain.get_all_goals())
        self.assertEqual(expected_goals, actual_result)

    def test_get_all_goals_in_defined_order(self):
        expected_goals = [
            ResolveGoal(self.price_question, speaker.SYS),
            PerformGoal(self.buy_action),
            PerformGoal(self.always_preferred_action),
            PerformGoal(self.conditionally_preferred_action),
            PerformGoal(self.accommodate_unrestricted_action),
            PerformGoal(self.downdate_plan_true_action),
            PerformGoal(self.downdate_plan_false_action),
            PerformGoal(self.instructional),
            PerformGoal(self.top_action),
            PerformGoal(self.up_action),
        ]
        actual_result = self.domain.get_all_goals_in_defined_order()
        self.assertEqual(expected_goals, actual_result)

    def test_get_dependent_question_returns_goal_issue_for_question_in_plan(self):
        self.assertEqual(self.price_question, self.domain.get_dependent_question(self.dest_city_question))

    def test_get_dependent_question_returns_none_for_goal_issue(self):
        self.assertEqual(None, self.domain.get_dependent_question(self.price_question))

    def test_iterate_plans(self):
        expected_goals = {
            ResolveGoal(self.price_question, speaker.SYS),
            PerformGoal(self.buy_action),
            PerformGoal(self.always_preferred_action),
            PerformGoal(self.conditionally_preferred_action),
            PerformGoal(self.up_action),
            PerformGoal(self.accommodate_unrestricted_action),
            PerformGoal(self.top_action),
            PerformGoal(self.downdate_plan_true_action),
            PerformGoal(self.downdate_plan_false_action),
            PerformGoal(self.instructional),
        }
        actual_goals = {goal for goal in self.domain.get_plan_goal_iterator()}
        self.assertEqual(expected_goals, actual_goals)

    def test_get_downdate_conditions_for_goal_with_downdate_condition(self):
        action_with_downdate_conditions = self.buy_action
        self.assertEqual([self.condition],
                         list(self.domain.get_downdate_conditions(PerformGoal(action_with_downdate_conditions))))

    def test_get_downdate_conditions_for_goal_without_downdate_conditions(self):
        action_without_downdate_condition = self.always_preferred_action
        self.assertEqual([], list(self.domain.get_downdate_conditions(PerformGoal(action_without_downdate_condition))))

    def test_get_downdate_conditions_returns_implicit_downdate_conditions_for_service_action_invocations_if_downdate_plan_is_true(
        self
    ):
        self.assertEqual([ServiceActionTerminatedProposition("mockup_ontology", "mock_service_action")],
                         list(self.domain.get_downdate_conditions(PerformGoal(self.downdate_plan_true_action))))

    def test_get_downdate_conditions_doesnt_return_implicit_downdate_conditions_for_service_action_invocations_if_not_downdating_plan(
        self
    ):
        self.assertEqual([], list(self.domain.get_downdate_conditions(PerformGoal(self.downdate_plan_false_action))))

    def test_get_downdate_conditions_for_resolve_goal(self):
        resolve_goal = ResolveGoal(self.price_question, speaker.SYS)
        self.assertEqual([], list(self.domain.get_downdate_conditions(resolve_goal)))

    def test_get_postplan_for_goal_with_postplan(self):
        action_with_postplan = self.always_preferred_action
        postplan = self.domain.get_postplan(PerformGoal(action_with_postplan))
        expected_postplan = [Findout(self.domain_name, self.price_question)]
        self.assertEqual(expected_postplan, postplan)

    def test_get_postplan_for_goal_without_postplan(self):
        action_without_postplan = self.buy_action
        self.assertEqual([], self.domain.get_postplan(PerformGoal(action_without_postplan)))

    def test_get_superactions_for_goal_with_superactions(self):
        action_with_superactions = self.always_preferred_action
        expected_result = [self.buy_action]
        self.assertEqual(expected_result, self.domain.get_superactions(PerformGoal(action_with_superactions)))

    def test_get_superactions_for_goal_without_superactions(self):
        action_without_superactions = self.buy_action
        self.assertEqual([], self.domain.get_superactions(PerformGoal(action_without_superactions)))

    def test_is_depending_on_true_for_goal_issue_wrt_question_in_its_plan(self):
        goal_issue = self.price_question
        question_in_plan = self.dest_city_question
        self.assertTrue(self.domain.is_depending_on(goal_issue, question_in_plan))

    def test_is_depending_on_false_for_plan_question_wrt_its_goal_issue(self):
        goal_issue = self.price_question
        question_in_plan = self.dest_city_question
        self.assertFalse(self.domain.is_depending_on(question_in_plan, goal_issue))

    def test_is_depending_on_true_due_to_domain_declaration(self):
        self.assertTrue(self.domain.is_depending_on(self.dest_country_question, self.dest_city_question))

    def test_is_depending_on_false_for_inverse_domain_declaration(self):
        self.assertFalse(self.domain.is_depending_on(self.dest_city_question, self.dest_country_question))

    def test_is_depending_on_true_because_of_featurehood(self):
        self.assertTrue(self.domain.is_depending_on(self.dest_city_question, self.dest_city_type_question))

    def test_is_depending_on_robust_to_non_sortal_questions(self):
        city_prop_set = PropositionSet([
            self.proposition_dest_city_london,
            self.proposition_dest_city_paris,
        ])
        non_sortal_question = AltQuestion(city_prop_set)

        self.domain.is_depending_on(non_sortal_question, non_sortal_question)

    def test_get_resolving_answers(self):
        question = self.dest_city_question
        expected_answers = [self.proposition_dest_city_paris, self.proposition_dest_city_london]
        self.assertEqual(expected_answers, self.domain.get_resolving_answers(question))

    def test_get_question_verbalize_returns_true_by_default(self):
        question_with_verbalize_unspecified = self.price_question
        self.assertTrue(self.domain.get_verbalize(question_with_verbalize_unspecified))

    def test_get_question_verbalize_returns_false_when_specified_as_false(self):
        self.assertFalse(self.domain.get_verbalize(self.non_verbalized_question))

    def test_empty_top_plan_by_default(self):
        self.assertEqual(Plan([]), self.domain.get_plan(PerformGoal(self.top_action)))

    def test_multiple_plans_for_same_goal_yields_exception(self):
        invalid_plans = [
            {
                "goal": PerformGoal(self.top_action),
                "plan": Plan([])
            },
            {
                "goal": PerformGoal(self.top_action),
                "plan": Plan([])
            },
        ]

        with self.assertRaises(InvalidPlansException):
            Domain(self.DDD_NAME, self.domain_name, self.ontology, invalid_plans)

    def test_get_questions_in_plan_base_case(self):
        self.given_plan([
            Findout(self.domain_name, self.price_question),
            Raise(self.domain_name, self.dest_city_question)
        ])
        self.when_get_questions_in_plan()
        self.then_result_has_elements([self.price_question, self.dest_city_question, self.dest_city_type_question])

    def when_get_questions_in_plan(self):
        self._actual_result = [question for question in self.domain.get_questions_in_plan(self._plan)]

    def then_result_has_elements(self, expected_result):
        self.assertEqual(set(expected_result), set(self._actual_result))

    def given_plan(self, plan_contents):
        self._plan = Plan(plan_contents)

    def test_get_questions_in_plan_supports_if_then_else(self):
        self.given_plan([
            IfThenElse(
                "mock_condition", [Findout(self.domain_name, self.price_question)], [Bind(self.dest_city_question)]
            )
        ])
        self.when_get_questions_in_plan()
        self.then_result_has_elements([self.price_question, self.dest_city_question, self.dest_city_type_question])

    def test_get_questions_in_plan_includes_wh_question_via_featurehood_in_ontology(self):
        self.given_ontology(
            sorts={CustomSort(self.ontology_name, "housing"),
                   CustomSort(self.ontology_name, "housing_type")},
            predicates={
                self._create_predicate("selected_housing", CustomSort(self.ontology_name, "housing")),
                self._create_predicate(
                    "selected_housing_type",
                    CustomSort(self.ontology_name, "housing_type"),
                    feature_of_name="selected_housing"
                )
            }
        )
        self.given_domain()
        self.given_plan([Findout(self.domain_name, self.ontology.create_wh_question("selected_housing"))])
        self.when_get_questions_in_plan()
        self.then_result_has_elements([
            self.ontology.create_wh_question("selected_housing"),
            self.ontology.create_wh_question("selected_housing_type")
        ])

    def test_get_questions_in_plan_includes_wh_question_via_ask_features_parameter(self):
        self.given_ontology(
            sorts={CustomSort(self.ontology_name, "housing"),
                   CustomSort(self.ontology_name, "housing_type")},
            predicates={
                self._create_predicate("selected_housing", CustomSort(self.ontology_name, "housing")),
                self._create_predicate("selected_housing_type", CustomSort(self.ontology_name, "housing_type"))
            }
        )
        self.given_domain(
            parameters={
                self.ontology.create_wh_question("selected_housing"): {
                    "ask_features": [AskFeature("selected_housing_type")]
                }
            }
        )
        self.given_plan([Findout(self.domain_name, self.ontology.create_wh_question("selected_housing"))])
        self.when_get_questions_in_plan()
        self.then_result_has_elements([
            self.ontology.create_wh_question("selected_housing"),
            self.ontology.create_wh_question("selected_housing_type")
        ])

    def test_get_questions_in_plan_includes_kpq_via_ask_features_parameter(self):
        self.given_ontology(
            sorts={CustomSort(self.ontology_name, "housing"),
                   CustomSort(self.ontology_name, "housing_type")},
            predicates={
                self._create_predicate("selected_housing", CustomSort(self.ontology_name, "housing")),
                self._create_predicate("selected_housing_type", CustomSort(self.ontology_name, "housing_type"))
            }
        )
        self.given_domain(
            parameters={
                self.ontology.create_wh_question("selected_housing"): {
                    "ask_features": [AskFeature("selected_housing_type", kpq=True)]
                }
            }
        )
        self.given_plan([Findout(self.domain_name, self.ontology.create_wh_question("selected_housing"))])
        self.when_get_questions_in_plan()
        self.then_result_has_elements([
            self.ontology.create_wh_question("selected_housing"),
            self.ontology.create_wh_question("selected_housing_type"),
            KnowledgePreconditionQuestion(self.ontology.create_wh_question("selected_housing_type"))
        ])

    def given_domain(self, plans=None, default_questions=None, dependencies=None, parameters=None):
        if plans is None:
            plans = []
        if default_questions is None:
            default_questions = []
        if dependencies is None:
            dependencies = {}
        if parameters is None:
            parameters = {}

        self.domain = Domain(
            self.DDD_NAME,
            self.domain_name,
            self.ontology,
            plans=plans,
            default_questions=default_questions,
            dependencies=dependencies,
            parameters=parameters
        )

        domain_as_json = self.domain.as_json()
        json_parser = JSONDomainParser(self.ontology)
        self.domain = json_parser.parse(domain_as_json)

    def given_parameters(self, parameters):
        self.domain.parameters = parameters

    def given_ontology(self, sorts=None, predicates=None, individuals=None, actions=None):
        if sorts is None:
            sorts = {}
        if predicates is None:
            predicates = {}
        if individuals is None:
            individuals = {}
        if actions is None:
            actions = {}
        self.ontology = Ontology(self.ontology_name, sorts, predicates, individuals, self.actions)

    def test_get_user_targeted_actions(self):
        expected_action_names = ["user_targeted_action"]
        actual_result = self.domain.get_names_of_user_targeted_actions()
        self.assertEqual(set(expected_action_names), set(actual_result))


class DominatesTests(LibTestCase):
    def setUp(self):
        self.ontology_name = "mockup_ontology"
        actions = {
            "dominating_action", "dominated_action", "super_dominated_action", "non_dominated_action",
            "planless_action", "action_w_predicate_proposition_alt_question"
        }
        sorts = set()
        predicates = {
            self._create_predicate("dominated_issue", RealSort()),
            self._create_predicate("non_dominated_issue", RealSort())
        }
        individuals = {}
        self.ontology_name = "mockup_ontology"
        self.ontology = Ontology(self.ontology_name, sorts, predicates, individuals, actions)

        self.domain_name = "mockup_domain"
        self.DDD_NAME = "mock_ddd"

        for action_name in actions:
            action = self.ontology.create_action(action_name)
            setattr(self, action_name, action)

        for question_name in ["dominated_issue", "non_dominated_issue"]:
            question = self.ontology.create_wh_question(question_name)
            setattr(self, question_name, question)

        plans = [{
            "goal": PerformGoal(self.dominating_action),
            "plan": [
                self._findout_with_alts([
                    GoalProposition(PerformGoal(self.dominated_action)),
                    GoalProposition(ResolveGoal(self.dominated_issue, speaker.SYS))
                ])
            ]
        }, {
            "goal": PerformGoal(self.dominated_action),
            "plan": [self._findout_with_alts([GoalProposition(PerformGoal(self.super_dominated_action))])]
        }, {
            "goal": PerformGoal(self.super_dominated_action),
            "plan": []
        }, {
            "goal": ResolveGoal(self.dominated_issue, speaker.SYS),
            "plan": []
        }, {
            "goal": PerformGoal(self.non_dominated_action),
            "plan": []
        }, {
            "goal": PerformGoal(self.action_w_predicate_proposition_alt_question),
            "plan": [
                self._findout_with_alts([
                    PredicateProposition(
                        self._create_predicate("a_predicate", RealSort()),
                        self.ontology.create_individual(5.0, RealSort()), self.ontology
                    )
                ])
            ]
        }]

        self.domain = Domain(self.DDD_NAME, self.domain_name, self.ontology, plans=plans)

    def test_dominates_true_for_dominated_action(self):
        self.assertTrue(self.domain.dominates(PerformGoal(self.dominating_action), PerformGoal(self.dominated_action)))

    def test_dominates_true_for_super_dominated_action(self):
        self.assertTrue(
            self.domain.dominates(PerformGoal(self.dominating_action), PerformGoal(self.super_dominated_action))
        )

    def test_dominates_true_for_dominated_issue(self):
        self.assertTrue(
            self.domain.dominates(PerformGoal(self.dominating_action), ResolveGoal(self.dominated_issue, speaker.SYS))
        )

    def test_dominates_false_for_non_dominated_action(self):
        self.assertFalse(
            self.domain.dominates(PerformGoal(self.dominating_action), PerformGoal(self.non_dominated_action))
        )

    def test_dominates_false_for_non_dominated_issue(self):
        self.assertFalse(
            self.domain.dominates(
                PerformGoal(self.dominating_action), ResolveGoal(self.non_dominated_issue, speaker.SYS)
            )
        )

    def test_dominates_false_for_action_without_plan(self):
        self.assertFalse(self.domain.dominates(PerformGoal(self.planless_action), PerformGoal(self.dominated_action)))

    def test_dominates_false_for_action_w_predicate_proposition_alt_question(self):
        self.assertFalse(
            self.domain.dominates(
                PerformGoal(self.action_w_predicate_proposition_alt_question), PerformGoal(self.non_dominated_action)
            )
        )

    def _findout_with_alts(self, alts):
        return Findout(self.domain_name, AltQuestion(PropositionSet(alts)))



#!/usr/bin/env python

import argparse
import contextlib
import os
import sys
import warnings
import unittest
import logging
import time
import random
import signal

import structlog
import json

from requests.exceptions import MissingSchema

from tala.cli import console_formatting
from tala.cli.tdm.tdm_cli import TDMCLI
from tala.config import BackendConfig, DddConfig, DeploymentsConfig, BackendConfigNotFoundException, \
    DddConfigNotFoundException, DeploymentsConfigNotFoundException
from tala.ddd.maker.ddd_maker import DddMaker
from tala import installed_version
from tala.log.logger import configure_stdout_logging, configure_file_logging
from tala.log.interaction_test_extractor import InteractionTestExtractor
from tala.testing.interactions.file import InteractionTestingFile
from tala.testing.interactions.testloader import InteractionTestingLoader
from tala.testing.interactions.result import InteractionTestResult
from tala.utils.tdm_client import TDMClient
from tala.testing.endurance.runner import EnduranceTestRunner
from tala.cli.argument_parser import add_common_backend_arguments, add_shared_frontend_and_backend_arguments
from tala.log.formats import VALID_TIS_LOGGING_FORMATS, TIS_LOGGING_AUTO, VALID_TIS_UPDATE_FORMATS, TIS_LOGGING_FULL
from tala.service.constants import DEFAULT_PORT
from tala import domain_orchestration


class UnexpectedDDDException(Exception):
    pass


class ConfigAlreadyExistsException(Exception):
    pass


class ConfigNotFoundException(Exception):
    pass


class InvalidArgumentException(Exception):
    pass


def create_ddd(args):
    DddMaker(args.name, use_rgl=False, target_dir=args.target_dir).make()


def create_backend_config(args):
    if os.path.exists(args.filename):
        raise ConfigAlreadyExistsException(
            f"Expected to be able to create backend config file '{args.filename}' but it already exists."
        )
    BackendConfig().write_default_config(args.ddd, args.filename)


def configure_structlog_and_create_logger(args):
    if args.log_to_stdout:
        configure_stdout_logging(args.log_level)
    else:
        configure_file_logging(level=args.log_level)
    return structlog.get_logger("tala")


def create_orchestrated_domain_bundle(args):
    odb_as_json = domain_orchestration.create_odb_as_json(
        args.config, args.overridden_ddd_config_paths, args.should_greet, args.logged_tis_format,
        args.tis_update_format, args.log_to_stdout, args.log_level
    )
    odb = domain_orchestration.OrchestratedDomainBundle(odb_as_json).as_json()
    _store_odb(odb, args.output)


def _store_odb(odb, output_dir):
    print(f"storing orchestrated domain bundle in file '{output_dir}'")
    file = open(output_dir, 'w')
    file.write(json.dumps(odb))
    file.close()


def create_ddd_config(args):
    if os.path.exists(args.filename):
        raise ConfigAlreadyExistsException(
            f"Expected to be able to create DDD config file '{args.filename}' but it already exists."
        )
    DddConfig().write_default_config(path=args.filename, use_rgl=False)


def create_deployments_config(args):
    if os.path.exists(args.filename):
        raise ConfigAlreadyExistsException(
            f"Expected to be able to create deployments config file '{args.filename}' but it already exists."
        )
    DeploymentsConfig().write_default_config(args.filename)


@contextlib.contextmanager
def _config_exception_handling():
    def generate_message(name, command_name, config):
        return f"Expected {name} config '{config}' to exist but it was not found. To create it, run " \
               f"'tala {command_name} --filename {config}'."

    try:
        yield
    except BackendConfigNotFoundException as exception:
        message = generate_message("backend", "create-backend-config", exception.config_path)
        raise ConfigNotFoundException(message)
    except DddConfigNotFoundException as exception:
        message = generate_message("DDD", "create-ddd-config", exception.config_path)
        raise ConfigNotFoundException(message)
    except DeploymentsConfigNotFoundException as exception:
        message = generate_message("deployments", "create-deployments-config", exception.config_path)
        raise ConfigNotFoundException(message)


def version(args):
    print(installed_version.version)


def interact(args):
    config = DeploymentsConfig(args.deployments_config)
    url = config.get_url(args.environment_or_url)
    resume_session = args.resume_session
    tdm_cli = TDMCLI(url, resume_session)

    try:
        tdm_cli.run()
    except (KeyboardInterrupt, EOFError):
        tdm_cli.stop()
    except MissingSchema:
        environments = list(config.read().keys())
        print(f"Expected a URL or one of the known environments {environments} but got '{url}'")


def test_interactions(args):
    def _check_health(url):
        tdm = TDMClient(url)
        print(f"Connecting to {url}...", end="")
        sys.stdout.flush()
        try:
            tdm.wait_to_start()
            print("done.")
        except BaseException:
            print("failed.")
            raise
        sys.stdout.flush()

    def _test_suites(selected_tests, test_paths, url):
        test_files = [InteractionTestingFile.from_path(test_path) for test_path in test_paths]
        return _test_suites_from_files(selected_tests, test_files, url)

    def _test_suites_from_files(selected_tests, test_files, url):
        suites = []
        for file in test_files:
            test_loader = InteractionTestingLoader(url)
            suite = test_loader.load_interaction_tests(file, selected_tests)
            suites.append(suite)
        return suites

    def _run_tests_from_suites(suites):
        results = []
        runner = unittest.TextTestRunner(resultclass=InteractionTestResult)
        successful = False
        for suite in suites:
            print(f"Running interaction tests from {suite.filename}")
            result = runner.run(suite)
            results.append(result)
            successful = all([result.wasSuccessful() for result in results])
        return successful

    configure_stdout_logging(args.log_level)
    config = DeploymentsConfig(args.deployments_config)
    url = config.get_url(args.environment_or_url)
    _check_health(url)
    suites = _test_suites(args.selected_tests, args.tests_filenames, url)
    unittest.installHandler()
    successful = _run_tests_from_suites(suites)
    if not successful:
        sys.exit(-1)


def endurancetest(args):
    def on_terminate_signal(*args, **kwargs):
        test_runner.stop()

    configure_stdout_logging(args.log_level)
    config = DeploymentsConfig(args.deployments_config)
    url = config.get_url(args.environment_or_url)

    if args.seed:
        seed = args.seed
    else:
        seed = hash(time.time())
    random.seed(seed)
    print(f"Running endurance test with random seed {seed}")

    test_runner = EnduranceTestRunner(args.tests_filenames, args.duration, url)

    unittest.installHandler()

    signal.signal(signal.SIGINT, on_terminate_signal)
    signal.signal(signal.SIGTERM, on_terminate_signal)

    test_runner.run()


def extract(args):
    extractor = InteractionTestExtractor.from_args(
        args.log, verbose=args.verbose, full=args.full, semantic=args.semantic
    )
    extractor.run()


def add_create_ddd_subparser(subparsers):
    parser = subparsers.add_parser("create-ddd", help="create a new DDD")
    parser.set_defaults(func=create_ddd)
    parser.add_argument("name", help="Name of the DDD, e.g. my_ddd")
    parser.add_argument(
        "--target-dir", default=".", metavar="DIR", help="target directory, will be created if it doesn't exist"
    )


def add_create_odb_subparser(subparsers):
    parser = subparsers.add_parser("create-odb", help="create a new ODB")
    parser.set_defaults(func=create_orchestrated_domain_bundle)

    backend_group = parser.add_argument_group("backend")
    backend_group.add_argument("output", help="Directory where the ODB file will be stored")
    backend_group.add_argument("--should-greet", action="store_true", default=False)
    add_common_backend_arguments(backend_group)
    add_shared_frontend_and_backend_arguments(backend_group)
    logging_group = parser.add_argument_group("logging")
    logging_group.add_argument(
        "--log-to-stdout",
        action="store_true",
        default=False,
        help="If enabled, logging is done to stdout instead of to a file"
    )
    logging_group.add_argument("--port", type=int, default=DEFAULT_PORT, help="listen for connections on this port")
    choices = [TIS_LOGGING_AUTO] + VALID_TIS_LOGGING_FORMATS
    logging_group.add_argument(
        "--tis-format",
        "-tf",
        dest="logged_tis_format",
        choices=choices,
        default=TIS_LOGGING_AUTO,
        help="log TIS on this format"
    )
    logging_group.add_argument(
        "--tis-update-format",
        choices=VALID_TIS_UPDATE_FORMATS,
        default=TIS_LOGGING_FULL,
        help="log TIS updates on this format"
    )


def add_create_backend_config_subparser(subparsers):
    parser = subparsers.add_parser("create-backend-config", help="create a backend config")
    parser.set_defaults(func=create_backend_config)
    parser.add_argument("ddd", help="name of the active DDD, e.g. my_ddd")
    parser.add_argument(
        "--filename",
        default=BackendConfig.default_name(),
        metavar="NAME",
        help=f"filename of the backend config, e.g. {BackendConfig.default_name()}"
    )


def add_create_ddd_config_subparser(subparsers):
    parser = subparsers.add_parser("create-ddd-config", help="create a DDD config")
    parser.set_defaults(func=create_ddd_config)
    parser.add_argument(
        "--filename",
        default=DddConfig.default_name(),
        metavar="NAME",
        help=f"filename of the DDD config, e.g. {DddConfig.default_name()}"
    )


def add_create_deployments_config_subparser(subparsers):
    parser = subparsers.add_parser("create-deployments-config", help="create a deployments config")
    parser.set_defaults(func=create_deployments_config)
    parser.add_argument(
        "--filename",
        default=DeploymentsConfig.default_name(),
        metavar="NAME",
        help=f"filename of the deployments config, e.g. {DeploymentsConfig.default_name()}"
    )


def add_version_subparser(subparsers):
    parser = subparsers.add_parser("version", help="print the Tala version")
    parser.set_defaults(func=version)


def add_deployment_config_arguments(parser):
    parser.add_argument(
        "environment_or_url",
        help="this is either an environment, e.g. 'dev', pointing to a url in the deployments config; "
        "alternatively, if not an environment, this is considered a url in itself; "
        "the url should point to a TDM deployment, e.g. 'https://my-deployment.ddd.tala.cloud:9090/interact'"
    )
    parser.add_argument(
        "--config",
        dest="deployments_config",
        default=None,
        help=f"override the default deployments config {DeploymentsConfig.default_name()!r}"
    )
    parser.add_argument(
        "--resume-session",
        dest="resume_session",
        default=None,
        const="tala-session.state",
        nargs="?",
        help="Session object to be resumed. If parameter omitted, tala will resume latest session."
    )


def add_interact_subparser(subparsers):
    parser = subparsers.add_parser("interact", help="start an interactive chat with a deployed DDD")
    add_deployment_config_arguments(parser)
    parser.set_defaults(func=interact)


def _add_test_arguments(parser):
    add_deployment_config_arguments(parser)
    parser.add_argument(dest="tests_filenames", nargs="+", metavar="TEST-FILE", help="specify DDD test files")
    parser.add_argument("-t", dest="selected_tests", nargs="+", default=[], metavar="TEST", help="select test by name")
    parser.add_argument(
        "-l", "--log-level", nargs="?", default=logging.WARNING, metavar="LOG-LEVEL", help="select logging level"
    )


def add_test_subparser(subparsers):
    parser = subparsers.add_parser("test", help="run interaction tests")
    _add_test_arguments(parser)
    parser.set_defaults(func=test_interactions)


def add_endurancetest_subparser(subparsers):
    parser = subparsers.add_parser(
        "endurancetest",
        help="run the provided interaction tests in a random sequence on a single backend until they fail, error or "
        "until the duration elapses. The tests succeed when the duration has elapsed and no failure or error has "
        "been found."
    )
    _add_test_arguments(parser)
    parser.add_argument(
        "--duration",
        default=0,
        type=int,
        metavar="SECONDS",
        help="run until the duration has elapsed; if duration is 0, run forever"
    )
    parser.add_argument("--seed", type=int, metavar="INTEGER", help="random seed")
    parser.set_defaults(func=endurancetest)


def add_extract_subparser(subparsers):
    parser = subparsers.add_parser("extract", help="extract interaction test from a TDM log file")
    parser.add_argument("-v", "--verbose", action="store_true", dest="verbose", help="print verbose information")
    parser.add_argument("-f", "--full", action="store_true", help="print full interpretation information")
    parser.add_argument("-s", "--semantic", action="store_true", help="print semantic information")
    parser.set_defaults(func=extract)
    parser.add_argument("log", help="the TDM session log to extract from")


def format_warnings():
    def warning_on_one_line(message, category, _filename, _lineno, _file=None, _line=None):
        string = f"{category.__name__}: {message}\n"
        return console_formatting.bold(string)

    warnings.formatwarning = warning_on_one_line


def show_deprecation_warnings():
    warnings.simplefilter("always", category=DeprecationWarning)


def main(args=None):
    format_warnings()
    show_deprecation_warnings()
    root_parser = argparse.ArgumentParser(description="Use the Tala SDK for the Talkamatic Dialogue Manager (TDM).")
    subparsers = root_parser.add_subparsers()
    add_create_ddd_subparser(subparsers)
    add_create_odb_subparser(subparsers)
    add_create_backend_config_subparser(subparsers)
    add_create_ddd_config_subparser(subparsers)
    add_create_deployments_config_subparser(subparsers)
    add_version_subparser(subparsers)
    add_interact_subparser(subparsers)
    add_test_subparser(subparsers)
    add_endurancetest_subparser(subparsers)
    add_extract_subparser(subparsers)

    parsed_args = root_parser.parse_args(args)
    with _config_exception_handling():
        parsed_args.func(parsed_args)


if __name__ == "__main__":
    main()



import uuid
import re
import json
import requests

from tala.model.interpretation import Interpretation
from tala.model.input_hypothesis import InputHypothesis
from tala.model.common import Modality
from tala.model.user_move import UserMove, DDDSpecificUserMove
from tala.utils.tdm_client import TDMClient

from tala.testing.interaction_test_noveau.comparison import StringComparison, MoveComparison

SPEAKER = "speaker"
USER = "user"
SYSTEM = "system"
TEST_NAME = "name"
MOVE_CONTENT = "move_content"
INTERPRETATIONS = "interpretations"
SPEECH_CONTENT = "speech_content"
OUTPUT = "output"
MOVES = "moves"
SESSION = "session"
EXPECTED_PASSIVITY = "expected_passivity"
UTTERANCE = "utterance"

TDM_PROTOCOL_VERSION = "3.4"

DEFAULT_DEVICE_ID = "interaction-tester"


class OutputBuffer:
    def __init__(self):
        self._content_lines = []

    def add(self, new_content):
        self._content_lines.append(new_content)

    def __str__(self):
        output = ""
        for line in self._content_lines:
            output += line + "\n"
        return output


class InteractionTester():
    def __init__(self, port, device_id=DEFAULT_DEVICE_ID):
        self._session_id = f"interaction-tester-session-{str(uuid.uuid4())}"
        self._device_id = device_id
        self._initialize_session_object()
        self._port = port

    def _initialize_session_object(self):
        self._session_data = {"device_id": self._device_id, "session_id": self._session_id}

    def start_session(self, session_data=None):
        if session_data:
            self._session_data = session_data
        self._latest_response = self._client.start_session(self._session_data)
        self._session_data = self._latest_response[SESSION]

    def run_testcase(self, case):
        self._initialize_testcase(case)
        self.start_session()
        self.line_number = 0
        success = True
        self._previous_entry_type = None
        for entry in case["interaction"]:
            self.line_number += 1
            if entry[SPEAKER] == USER:
                success = self._do_user_turn(entry)
            elif entry[SPEAKER] == SYSTEM:
                if self._previous_entry_type == SYSTEM:
                    raise Exception(f"Two consecutive entries define '{SYSTEM}' input")
                success = self._do_system_turn(entry)
            if not success:
                return self._create_response(self._result)
            self._previous_entry_type = entry[SPEAKER]
        self._buffer_output('=== End interaction test ===')
        self._result = {"success": True}
        return self._create_response(self._result)

    def _initialize_testcase(self, testcase):
        self._output_buffer = OutputBuffer()
        self._ddd_name = testcase["target_ddd"]
        url = self._patch_url_with_port(testcase["url"])
        self._client = TDMClient(url)
        self._test_name = testcase["name"]
        self._buffer_output(f'\n=== Begin interaction test "{self._test_name}" ===')

    def _patch_url_with_port(self, url):
        if self._port:
            new_url = re.sub(r"^(https?:[^:]*):\d+/(.+)", rf"\1:{self._port}/\2", url)
            return new_url
        return url

    def _buffer_output(self, output):
        self._output_buffer.add(output)

    def _print_buffer(self):
        print(str(self._output_buffer))

    def check_for_consecutive_speaker(self, speaker):
        if self._previous_entry_type == speaker:
            raise Exception(f"Two consecutive entries define '{speaker}' input")

    def _do_user_turn(self, user_entry):
        def create_interpretation(moves, utterance_content=""):
            return Interpretation([self._create_user_move(move) for move in moves], Modality.OTHER, utterance_content)

        def create_interpretations_from_dicts(dict_list, utterance):
            interpretations = []
            for entry in dict_list:
                interpretations.append(
                    Interpretation([create_user_move(move) for move in entry["moves"]], entry["modality"], utterance)
                )
            return interpretations

        def create_user_move(move_dict):
            return DDDSpecificUserMove(
                move_dict["ddd"], move_dict["semantic_expression"], move_dict["perception_confidence"],
                move_dict["understanding_confidence"]
            )

        def interpretations_as_json(interpretations):
            return json.dumps([interpretation.as_json() for interpretation in interpretations])

        self.check_for_consecutive_speaker(USER)

        if MOVE_CONTENT in user_entry:
            moves = user_entry[MOVE_CONTENT]
            self._buffer_output(f"U> {json.dumps(moves)}")
            utterance = user_entry.get("utterance", "")
            interpretation = create_interpretation(moves, utterance)
            self._request_semantic_input([interpretation])
        elif INTERPRETATIONS in user_entry:
            utterance = user_entry.get("utterance", "")
            interpretations = create_interpretations_from_dicts(user_entry[INTERPRETATIONS], utterance)
            entities = user_entry.get("entities", [])
            self._buffer_output(f"U> {utterance if utterance else interpretations_as_json(interpretations)}")
            self._request_semantic_input(interpretations, entities=entities)
        elif EXPECTED_PASSIVITY in user_entry:
            expected_passivity = user_entry[EXPECTED_PASSIVITY]
            self._buffer_output(f"U> {{expected_passivity={expected_passivity}}}")
            if self._passivity_mismatch(user_entry[EXPECTED_PASSIVITY]):
                self._create_passivity_mismatch_description(user_entry[EXPECTED_PASSIVITY])
                return False
            self._request_passivity()
        elif SPEECH_CONTENT in user_entry:
            utterance = user_entry.get(SPEECH_CONTENT)
            self._buffer_output(f"U> {utterance}")
            self._request_speech_input(utterance)
        else:
            raise Exception("Nothing to do in user entry:", user_entry)
        return True

    def _request_semantic_input(self, interpretations, entities=None):
        self._latest_response = self._client.request_semantic_input(interpretations, self._session_data, entities)
        self._update_session_data()

    def _request_passivity(self):
        self._latest_response = self._client.request_passivity(self._session_data)
        self._update_session_data()

    def _request_speech_input(self, utterance):
        hypotheses = [InputHypothesis(utterance, 1.0)]
        self._latest_response = self._client.request_speech_input(hypotheses, self._session_data)
        self._update_session_data()

    def _update_session_data(self):
        self._session_data = self._latest_response["session"]

    def _create_user_move(self, move):
        if self._ddd_name:
            return DDDSpecificUserMove(self._ddd_name, move, 1.0, 1.0)
        return UserMove(move, 1.0, 1.0)

    def _do_system_turn(self, system_entry):
        self.check_for_consecutive_speaker(SYSTEM)

        while self._is_request_for_service_invocation():
            self._make_service_request_and_create_tdm_request_with_service_invocation_result()
        else:
            if EXPECTED_PASSIVITY in system_entry and self._passivity_mismatch(system_entry[EXPECTED_PASSIVITY]):
                return self._create_passivity_mismatch_description(system_entry[EXPECTED_PASSIVITY])
            if MOVE_CONTENT in system_entry:
                return self._assert_system_moves_are_matched_by(system_entry[MOVE_CONTENT])
            if SPEECH_CONTENT in system_entry:
                return self._assert_system_utterance_is_matched_by(system_entry[SPEECH_CONTENT])

    def _is_request_for_service_invocation(self):
        if self._is_request_for_service_query_invocation():
            return True
        elif self._is_request_for_service_validation_invocation():
            return True
        elif self._is_request_for_service_action_invocation():
            return True

    def _is_request_for_service_query_invocation(self):
        if "predicate" in self._latest_response:
            return True

    def _is_request_for_service_validation_invocation(self):
        if "validator_name" in self._latest_response:
            return True

    def _is_request_for_service_action_invocation(self):
        if "action_name" in self._latest_response:
            return True

    def _make_service_request_and_create_tdm_request_with_service_invocation_result(self):
        if self._is_request_for_service_query_invocation():
            self._make_service_request_and_create_tdm_request_with_service_query_invocation_result()
        elif self._is_request_for_service_validation_invocation():
            self._make_service_request_and_create_tdm_request_with_service_validator_invocation_result()
        elif self._is_request_for_service_action_invocation():
            self._make_service_request_and_create_tdm_request_with_service_action_invocation_result()

    def _make_service_request_and_create_tdm_request_with_service_query_invocation_result(self):
        predicate = self._latest_response["predicate"]
        url = self._latest_response["url"]
        parameters = self._latest_response["parameters"]
        session = self._latest_response["session"]

        query_results = self._make_query_to_http_service(predicate, url, parameters, session)
        self._make_results_request_for_dme(
            "query_results", {
                "predicate": predicate,
                "url": url,
                "parameters": parameters,
                "ddd_name": self._latest_response["ddd_name"],
                "results": query_results["data"]["result"],
                "earlier_results": self._latest_response["earlier_results"]
            }
        )

    def _make_results_request_for_dme(self, type_, results):
        request = {
            "version": TDM_PROTOCOL_VERSION,
            "request": self._latest_response["original_request"],
            "session": self._latest_response["session"]
        }
        request["request"][type_] = results
        self._latest_response = self._client.make_request(request)

    def _make_query_to_http_service(self, name, url, parameters, session):
        data = {
            "session": session,
            "request": {
                "type": "query",
                "name": name,
                "parameters": parameters,
            }
        }
        response = requests.post(url, data=json.dumps(data), headers={"Content-type": "application/json"})
        response_dict = json.loads(response.text)
        return response_dict

    def _make_service_request_and_create_tdm_request_with_service_validator_invocation_result(self):
        validator_name = self._latest_response["validator_name"]
        url = self._latest_response["url"]
        parameters = self._latest_response["parameters"]
        session = self._latest_response["session"]

        query_results = self._validate_in_http_service(validator_name, url, parameters, session)
        self._make_results_request_for_dme(
            "validation_results", {
                "validator_name": validator_name,
                "url": url,
                "parameters": parameters,
                "ddd_name": self._latest_response["ddd_name"],
                "result": query_results,
                "earlier_results": self._latest_response["earlier_results"]
            }
        )

    def _validate_in_http_service(self, name, url, parameters, session):
        data = {
            "session": session,
            "request": {
                "type": "validator",
                "name": name,
                "parameters": parameters,
            }
        }
        response = requests.post(url, data=json.dumps(data), headers={"Content-type": "application/json"})
        response_dict = json.loads(response.text)
        return response_dict

    def _make_service_request_and_create_tdm_request_with_service_action_invocation_result(self):
        action_name = self._latest_response["action_name"]
        url = self._latest_response["url"]
        parameters = self._latest_response["parameters"]
        session = self._latest_response["session"]

        action_result = self._perform_action_in_http_service(action_name, url, parameters, session)
        self._make_results_request_for_dme(
            "action_results", {
                "action_name": action_name,
                "url": url,
                "parameters": parameters,
                "ddd_name": self._latest_response["ddd_name"],
                "result": action_result,
                "earlier_results": self._latest_response["earlier_results"]
            }
        )

    def _perform_action_in_http_service(self, name, url, parameters, session):
        data = {
            "session": session,
            "request": {
                "type": "action",
                "name": name,
                "parameters": parameters,
            }
        }
        response = requests.post(url, data=json.dumps(data), headers={"Content-type": "application/json"})
        response_dict = json.loads(response.text)
        return response_dict

    def _passivity_mismatch(self, expected_passivity_value):
        actual_value = self._latest_response[OUTPUT].get(EXPECTED_PASSIVITY)
        if not actual_value and actual_value != 0.0:
            return True
        if expected_passivity_value is True:
            return False
        if expected_passivity_value == actual_value:
            return False
        return True

    def _create_passivity_mismatch_description(self, expected_passivity_value):
        actual_value = self._latest_response[OUTPUT].get(EXPECTED_PASSIVITY, False)
        if actual_value is None:
            if expected_passivity_value is True:
                self._result = {
                    "success": False,
                    "failure_description": "Expected an expected_passivity, but none was set."
                }
            else:
                self._result = {
                    "success": False,
                    "failure_description": f"Expected expected_passivity={expected_passivity_value}, " +
                    "but no expected_passivity was set."
                }
        else:
            self._result = {
                "success": False,
                "failure_description": f"Expected expected_passivity={expected_passivity_value}, " +
                f"but actual expected_passivity was {actual_value}."
            }

    def _assert_system_moves_are_matched_by(self, expected_move_content):
        assert OUTPUT in self._latest_response, f"No {OUTPUT} in {self._latest_response}"
        actual_move_content = self._latest_response[OUTPUT][MOVES]
        comparison = MoveComparison(actual_move_content, expected_move_content)
        if not comparison.match():
            self._result = {"success": False, "failure_description": comparison.mismatch_description()}
            self._buffer_output(comparison.mismatch_description())
            return False
        self._buffer_output(f"S> {json.dumps(actual_move_content)}")
        return True

    def _assert_system_utterance_is_matched_by(self, expected_speech_content):
        actual_utterance_content = self._latest_response[OUTPUT][UTTERANCE]
        comparison = StringComparison(actual_utterance_content, expected_speech_content)
        if not comparison.match():
            self._result = {"success": False, "failure_description": comparison.mismatch_description()}
            self._buffer_output(comparison.mismatch_description())
            return False
        self._buffer_output(f"S> {actual_utterance_content}")
        return True

    def _create_response(self, response):
        response["transcript"] = str(self._output_buffer)
        return response



class Image(object):
    def __init__(self, url):
        self._url = url

    @property
    def url(self):
        return self._url

    def __eq__(self, other):
        try:
            return other.url == self.url
        except AttributeError:
            return False

    def __ne__(self, other):
        return not (other == self)

    def __str__(self):
        return 'image("%s")' % self.url

    def __repr__(self):
        return str(self)

    def __hash__(self):
        return 17 * hash(str(self)) + 7 * hash(self.__class__.__name__)



import warnings

from tala.model.semantic_object import OntologySpecificSemanticObject, SemanticObject
from tala.utils.as_semantic_expression import AsSemanticExpressionMixin


class LambdaAbstractedProposition():
    LAMBDA_ABSTRACTED_PREDICATE_PROPOSITION = "LambdaAbstractedPredicateProposition"
    LAMBDA_ABSTRACTED_IMPLICATION_PROPOSITION_FOR_CONSEQUENT = "LambdaAbstractedImplicationPropositionForConsequent"
    LAMBDA_ABSTRACTED_GOAL_PROPOSITION = "LambdaAbstractedGoalProposition"

    def __init__(self, type_):
        self.type_ = type_

    def as_dict(self):
        result = {"semantic_object_type": "LambdaAbstractedProposition"}
        return super().as_dict() | result


class LambdaAbstractedPredicateProposition(
    LambdaAbstractedProposition, OntologySpecificSemanticObject, AsSemanticExpressionMixin
):
    def __init__(self, predicate, ontology_name):
        OntologySpecificSemanticObject.__init__(self, ontology_name)
        LambdaAbstractedProposition.__init__(self, self.LAMBDA_ABSTRACTED_PREDICATE_PROPOSITION)
        self.predicate = predicate

    def is_lambda_abstracted_predicate_proposition(self):
        return True

    def __str__(self):
        variable = "X"
        return variable + "." + self.predicate.get_name() + "(" + variable + ")"

    def __eq__(self, other):
        if (isinstance(other, LambdaAbstractedPredicateProposition)):
            return self.predicate == other.predicate
        else:
            return False

    def __ne__(self, other):
        return not (self == other)

    def getPredicate(self):
        warnings.warn(
            "LambdaAbstractedPredicateProposition.getPredicate() is deprecated. "
            "Use LambdaAbstractedPredicateProposition.predicate instead.",
            DeprecationWarning,
            stacklevel=2
        )
        return self.predicate

    @property
    def sort(self):
        return self.predicate.sort

    def getSort(self):
        warnings.warn(
            "LambdaAbstractedPredicateProposition.getSort() is deprecated. "
            "Use LambdaAbstractedPredicateProposition.sort instead.",
            DeprecationWarning,
            stacklevel=2
        )
        return self.sort

    def __hash__(self):
        return hash((self.__class__.__name__, self.predicate))


class LambdaAbstractedImplicationPropositionForConsequent(LambdaAbstractedProposition, OntologySpecificSemanticObject):
    def __init__(self, antecedent, consequent_predicate, ontology_name):
        OntologySpecificSemanticObject.__init__(self, ontology_name)
        LambdaAbstractedProposition.__init__(self, self.LAMBDA_ABSTRACTED_IMPLICATION_PROPOSITION_FOR_CONSEQUENT)
        self._antecedent = antecedent
        self._consequent_predicate = consequent_predicate

    @property
    def antecedent(self):
        return self._antecedent

    @property
    def consequent_predicate(self):
        return self._consequent_predicate

    def is_lambda_abstracted_implication_proposition_for_consequent(self):
        return True

    def __str__(self):
        return "X.implies(%s, %s(X))" % (self._antecedent, self._consequent_predicate)

    def __eq__(self, other):
        try:
            return other.antecedent == self.antecedent and other.consequent_predicate == self.consequent_predicate
        except AttributeError:
            return False

    def __hash__(self):
        return hash((self.__class__.__name__, self.antecedent, self.consequent_predicate))


class LambdaAbstractedGoalProposition(LambdaAbstractedProposition, SemanticObject, AsSemanticExpressionMixin):
    def __init__(self):
        SemanticObject.__init__(self)
        LambdaAbstractedProposition.__init__(self, self.LAMBDA_ABSTRACTED_GOAL_PROPOSITION)

    def is_lambda_abstracted_goal_proposition(self):
        return True

    def __eq__(self, other):
        try:
            return other.type_ == self.LAMBDA_ABSTRACTED_GOAL_PROPOSITION
        except Exception:
            return False

    def __ne__(self, other):
        return not (self == other)

    def __str__(self):
        return "X.goal(X)"

    def __hash__(self):
        return hash(self.__class__.__name__)



import unittest

from tala.model.commitments import Commitments
from tala.model.proposition import ActionStatusProposition
from tala.model.action_status import Done
from tala.model.action import Action


class CommitmentsTests(unittest.TestCase):
    def setUp(self):
        self.commitments = Commitments()

    def test_as_dict(self):
        self.given_a_commitment(ActionStatusProposition(Action("some_action", "some_ontology"), Done()))
        self.when_stored_as_json_dict()
        self.then_stored_dict_is({
            'commitments_set': {
                'set': [{
                    '_type': 'ACTION_STATUS',
                    '_polarity': 'POS',
                    '_predicted': False,
                    '_confidence_estimates': None,
                    '_content': {
                        '_ontology_name': 'some_ontology',
                        'value': 'some_action',
                        'semantic_expression': 'some_action'
                    },
                    '_status': {
                        'action_status': 'done'
                    },
                    'semantic_expression': 'action_status(some_action, done)'
                }],
                'semantic_expression': '{action_status(some_action, done)}'
            },
            'meta_com': [{
                'key': {
                    '_type': 'ACTION_STATUS',
                    '_polarity': 'POS',
                    '_predicted': False,
                    '_confidence_estimates': None,
                    '_content': {
                        '_ontology_name': 'some_ontology',
                        'value': 'some_action',
                        'semantic_expression': 'some_action'
                    },
                    '_status': {
                        'action_status': 'done'
                    },
                    'semantic_expression': 'action_status(some_action, done)'
                },
                'value': [None, '-1']
            }],
            'pcom': {
                'set': [],
                'semantic_expression': '{}'
            },
            'meta_pcom': [],
            'commitment_lifespan': '30',
            'semantic_expression': '{action_status(some_action, done)}'
        })

    def given_a_commitment(self, commitment):
        self.commitments.add(commitment)

    def when_stored_as_json_dict(self):
        self.commitments_as_dict = self.commitments.as_json()

    def then_stored_dict_is(self, expected):
        self.assertEqual(expected, self.commitments_as_dict)

    def test_as_dict_when_commitment_is_removed(self):
        self.given_a_commitment(ActionStatusProposition(Action("some_action", "some_ontology"), Done()))
        self.given_commitment_was_removed(ActionStatusProposition(Action("some_action", "some_ontology"), Done()))
        self.when_stored_as_json_dict()
        self.then_stored_dict_is({
            'commitments_set': {
                'set': [],
                'semantic_expression': '{}'
            },
            'meta_com': [{
                'key': {
                    '_type': 'ACTION_STATUS',
                    '_polarity': 'POS',
                    '_predicted': False,
                    '_confidence_estimates': None,
                    '_content': {
                        '_ontology_name': 'some_ontology',
                        'value': 'some_action',
                        'semantic_expression': 'some_action'
                    },
                    '_status': {
                        'action_status': 'done'
                    },
                    'semantic_expression': 'action_status(some_action, done)'
                },
                'value': [None, '-1']
            }],
            'pcom': {
                'set': [{
                    '_type': 'ACTION_STATUS',
                    '_polarity': 'POS',
                    '_predicted': False,
                    '_confidence_estimates': None,
                    '_content': {
                        '_ontology_name': 'some_ontology',
                        'value': 'some_action',
                        'semantic_expression': 'some_action'
                    },
                    '_status': {
                        'action_status': 'done'
                    },
                    'semantic_expression': 'action_status(some_action, done)'
                }],
                'semantic_expression': '{action_status(some_action, done)}'
            },
            'meta_pcom': [{
                'key': {
                    '_type': 'ACTION_STATUS',
                    '_polarity': 'POS',
                    '_predicted': False,
                    '_confidence_estimates': None,
                    '_content': {
                        '_ontology_name': 'some_ontology',
                        'value': 'some_action',
                        'semantic_expression': 'some_action'
                    },
                    '_status': {
                        'action_status': 'done'
                    },
                    'semantic_expression': 'action_status(some_action, done)'
                },
                'value': [None, '-1']
            }],
            'commitment_lifespan': '30',
            'semantic_expression': '{}'
        })

    def given_commitment_was_removed(self, commitment):
        self.commitments.remove(commitment)

    def test_creation_from_dict_single_commitment(self):
        self.maxDiff = None
        self.given_commitments_as_dict({
            'commitments_set': {
                'set': [],
                'semantic_expression': '{}'
            },
            'meta_com': [{
                'key': {
                    '_type': 'ACTION_STATUS',
                    '_polarity': 'POS',
                    '_predicted': False,
                    '_confidence_estimates': None,
                    '_content': {
                        '_ontology_name': 'some_ontology',
                        'value': 'some_action',
                        'semantic_expression': 'some_action'
                    },
                    '_status': {
                        'action_status': 'done'
                    },
                    'semantic_expression': 'action_status(some_action, done)'
                },
                'value': [None, '-1']
            }],
            'pcom': {
                'set': [{
                    '_type': 'ACTION_STATUS',
                    '_polarity': 'POS',
                    '_predicted': False,
                    '_confidence_estimates': None,
                    '_content': {
                        '_ontology_name': 'some_ontology',
                        'value': 'some_action',
                        'semantic_expression': 'some_action'
                    },
                    '_status': {
                        'action_status': 'done'
                    },
                    'semantic_expression': 'action_status(some_action, done)'
                }],
                'semantic_expression': '{action_status(some_action, done)}'
            },
            'meta_pcom': [{
                'key': {
                    '_type': 'ACTION_STATUS',
                    '_polarity': 'POS',
                    '_predicted': False,
                    '_confidence_estimates': None,
                    '_content': {
                        '_ontology_name': 'some_ontology',
                        'value': 'some_action',
                        'semantic_expression': 'some_action'
                    },
                    '_status': {
                        'action_status': 'done'
                    },
                    'semantic_expression': 'action_status(some_action, done)'
                },
                'value': [None, '-1']
            }],
            'commitment_lifespan': '30',
            'semantic_expression': '{}'
        })
        self.given_created_commitments_from_input_dict()
        self.when_stored_as_json_dict()
        self.then_stored_dict_is(self.input_commitments_as_dict)

    def given_commitments_as_dict(self, commitments_as_dict):
        self.input_commitments_as_dict = commitments_as_dict

    def given_created_commitments_from_input_dict(self):
        self.commitments = Commitments.create_from_json(self.input_commitments_as_dict)

    def test_creation_from_dict_and_purge(self):
        self.maxDiff = None
        self.given_commitments_as_dict({
            'commitments_set': {
                'set': [],
                'semantic_expression': '{}'
            },
            'meta_com': [{
                'key': {
                    '_type': 'ACTION_STATUS',
                    '_polarity': 'POS',
                    '_predicted': False,
                    '_confidence_estimates': None,
                    '_content': {
                        '_ontology_name': 'some_ontology',
                        'value': 'some_action',
                        'semantic_expression': 'some_action'
                    },
                    '_status': {
                        'action_status': 'done'
                    },
                    'semantic_expression': 'action_status(some_action, done)'
                },
                'value': [None, '-1']
            }],
            'pcom': {
                'set': [{
                    '_type': 'ACTION_STATUS',
                    '_polarity': 'POS',
                    '_predicted': False,
                    '_confidence_estimates': None,
                    '_content': {
                        '_ontology_name': 'some_ontology',
                        'value': 'some_action',
                        'semantic_expression': 'some_action'
                    },
                    '_status': {
                        'action_status': 'done'
                    },
                    'semantic_expression': 'action_status(some_action, done)'
                }],
                'semantic_expression': '{action_status(some_action, done)}'
            },
            'meta_pcom': [{
                'key': {
                    '_type': 'ACTION_STATUS',
                    '_polarity': 'POS',
                    '_predicted': False,
                    '_confidence_estimates': None,
                    '_content': {
                        '_ontology_name': 'some_ontology',
                        'value': 'some_action',
                        'semantic_expression': 'some_action'
                    },
                    '_status': {
                        'action_status': 'done'
                    },
                    'semantic_expression': 'action_status(some_action, done)'
                },
                'value': [None, '-1']
            }],
            'commitment_lifespan': '30',
            'semantic_expression': '{}'
        })
        self.given_created_commitments_from_input_dict()
        self.when_purged()
        self.then_no_exceptions_are_raised()

    def when_purged(self):
        self.commitments.should_purge_previous_commitments(12)

    def then_no_exceptions_are_raised(self):
        self.assertTrue(True)



class ServiceActionOutcome():
    @property
    def is_successful(self):
        raise NotImplementedError("This property needs to be implemented in a subclass")

    def __eq__(self, other):
        try:
            return other.is_successful == self.is_successful
        except AttributeError:
            return False

    def __ne__(self, other):
        return not (self == other)

    def as_json(self):
        return {"service_action_outcome": None}


class Success(ServiceActionOutcome):
    @property
    def is_successful(self):
        return True

    def __repr__(self):
        return "%s()" % self.__class__.__name__

    def as_json(self):
        return {"service_action_outcome": True}


class SuccessfulServiceAction(Success):
    pass


class Failure(ServiceActionOutcome):
    def __init__(self, failure_reason):
        super().__init__()
        self._failure_reason = failure_reason

    @property
    def is_successful(self):
        return False

    @property
    def failure_reason(self):
        return self._failure_reason

    def as_json(self):
        return {"service_action_outcome": False, "failure_reason": self.failure_reason}

    def __eq__(self, other):
        return super().__eq__(other) and other.failure_reason == self.failure_reason

    def __str__(self):
        return "%s(%s)" % (self.__class__.__name__, self.failure_reason)

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.failure_reason)


class FailedServiceAction(Failure):
    def __init__(self, failure_reason):
        super().__init__(failure_reason)



import copy

from tala.model.action import Action
from tala.model.error import DomainError
from tala.model.goal import PerformGoal, ResolveGoal, PERFORM
from tala.model.question import KnowledgePreconditionQuestion
from tala.model import speaker
from tala.model.plan import Plan, InvalidPlansException
from tala.model import plan_item
from tala.model.proposition import PredicateProposition, ServiceActionTerminatedProposition
from tala.utils.as_json import AsJSONMixin, convert_to_json
from tala.utils.unique import unique

SEMANTIC_OBJECT_TYPE = "domain"


class DddDomain:
    pass


def dict_as_key_value_list(dict_):
    return [create_json_dict(key, value) for key, value in dict_.items()]


def create_json_dict(key, value):
    return {"key": convert_to_json(key), "value": convert_to_json(value)}


def queries_dict_as_list(dict_):
    def is_iterator(query_content):
        return "limit" in query_content

    queries_list = []
    for _question, content_dict in dict_.items():
        result_dict = {}
        if is_iterator(content_dict):
            result_dict = serialize_iterator(content_dict)
        else:
            for entry in content_dict:
                result_dict[entry] = convert_to_json(content_dict[entry])
        queries_list.append(result_dict)
    return queries_list


def serialize_iterator(iterator):
    serialized_iterator = {}
    for key, value in iterator.items():
        if key in ["for_enumeration", "for_random_enumeration"]:
            serialized_iterator[key] = [convert_to_json(item) for item in value]
        else:
            serialized_iterator[key] = value
    return serialized_iterator


class Domain(AsJSONMixin):

    def __init__(
            self, ddd_name, name, ontology, plans=None, default_questions=None, parameters=None,
            validators=None, dependencies=None, queries=None
    ):  # yapf: disable
        if plans is None:
            plans = []
        if default_questions is None:
            default_questions = []
        if parameters is None:
            parameters = {}
        if validators is None:
            validators = {}
        if dependencies is None:
            dependencies = {}
        if queries is None:
            queries = []

        self.ddd_name = ddd_name
        self.name = name
        self.ontology = ontology
        self.default_questions = default_questions
        self.parameters = parameters
        self.dependencies = dependencies
        self.plans = self._plan_list_to_dict_indexed_by_goal(plans)
        self.queries = self._query_list_to_dict_indexed_by_question(queries)
        self.validators = validators
        self._goals_in_defined_order = [plan["goal"] for plan in plans]
        self._add_top_plan_if_missing()
        self._add_up_plan()

    def as_dict(self):
        json = super(Domain, self).as_dict()
        json["ontology"] = "<skipped>"
        json["semantic_object_type"] = SEMANTIC_OBJECT_TYPE
        json["parameters"] = dict_as_key_value_list(self.parameters)
        json["dependencies"] = dict_as_key_value_list(self.dependencies)
        json["queries"] = queries_dict_as_list(self.queries)
        return json

    @property
    def goals(self):
        return list(self.plans.keys())

    def _has_top_plan(self):
        return any([
            plan for plan in list(self.plans.values()) if (
                plan["goal"].is_goal() and plan["goal"].type == PERFORM
                and plan["goal"].get_action().get_value() == "top"
            )
        ])  # noqa: E127

    def get_ontology(self):
        return self.ontology

    def __str__(self):
        return self.name

    def __eq__(self, other):
        return self.name == other.name and self.ddd_name == other.ddd_name

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash(self.name) * hash(self.ddd_name)

    def get_name(self):
        return self.name

    def action(self, action_name):
        return Action(action_name, self.ontology.get_name())

    def is_depending_on(self, dependent_issue, other):
        if dependent_issue in self.dependencies:
            if other in self.dependencies[dependent_issue]:
                return True
        if self._goal_issue_depends_on_question(dependent_issue, other):
            return True
        if self.is_feature_of(other, dependent_issue):
            return True
        return False

    def is_feature_of(self, feature_question, question):
        try:
            feature_question_predicate = feature_question.content.predicate
            question_predicate = question.content.predicate
        except AttributeError:
            return False
        return feature_question_predicate.is_feature_of(question_predicate)

    def _goal_issue_depends_on_question(self, goal_issue, question):
        resolve_goal = ResolveGoal(goal_issue, speaker.SYS)
        if self.has_goal(resolve_goal):
            plan = self.get_plan(resolve_goal)
            for item in plan:
                if item.type_ in plan_item.QUESTION_TYPES:
                    if question == item.content:
                        return True
        return False

    def get_dependent_question(self, question):
        for other_question in self.get_plan_questions():
            if self.is_depending_on(other_question, question):
                return other_question
        return None

    def has_goal(self, goal):
        return goal in self.plans

    def get_plan(self, goal):
        plan_info = self._get_plan_info(goal)
        return plan_info["plan"]

    def _get_plan_info(self, goal):
        if goal in self.plans:
            return self.plans[goal]
        else:
            raise DomainError("no plan for goal '%s' in domain '%s'" % (goal, self.get_name()))

    def dominates(self, supergoal, subgoal):
        if self.has_goal(supergoal):
            plan = self.get_plan(supergoal)
            for item in plan:
                if item.type_ in plan_item.QUESTION_TYPES:
                    question = item.content
                    if question.is_alt_question():
                        for proposition in question.content:
                            if proposition.is_goal_proposition():
                                goal = proposition.get_goal()
                                if goal == subgoal:
                                    return True
                                elif self.dominates(goal, subgoal):
                                    return True
        return False

    def _is_action_goal(self, goal):
        try:
            return goal.is_action()
        except AttributeError:
            return False

    def goal_is_preferred(self, goal):
        return self.goal_is_conditionally_preferred(goal, [])

    def goal_is_conditionally_preferred(self, goal, facts):
        condition = self.get_preferred(goal)
        if condition is True or condition in facts:
            return True

    def is_default_question(self, question):
        return question in self.default_questions

    def get_plan_questions(self):
        already_found = set()
        for goal in self.plans:
            plan = self.get_plan(goal)
            for question in self.get_questions_in_plan(plan):
                if question not in already_found:
                    already_found.add(question)
                    yield question

    def get_plan_goal_iterator(self):
        for goal in self.plans:
            yield goal

    def get_all_goals(self):
        return list(self.plans.keys())

    def get_all_goals_in_defined_order(self):
        return self._goals_in_defined_order

    def get_all_resolve_goals(self):
        return [x for x in self.plans if x.is_resolve_goal()]

    def get_downdate_conditions(self, goal):
        if self.has_goal(goal):
            for downdate_condition in self.get_goal_attribute(goal, 'postconds'):
                yield downdate_condition
            for downdate_condition in self._implicit_downdate_conditions_for_service_action_invocations(goal):
                yield downdate_condition
            for downdate_condition in self._implicit_downdate_conditions_for_handle_goal(goal):
                yield downdate_condition

    def _implicit_downdate_conditions_for_service_action_invocations(self, goal):
        plan = self.get_plan(goal)
        for item in plan:
            if item.type_ == plan_item.TYPE_INVOKE_SERVICE_ACTION and item.should_downdate_plan():
                yield ServiceActionTerminatedProposition(self.ontology.name, item.get_service_action())

    def _implicit_downdate_conditions_for_handle_goal(self, goal):
        if goal.is_handle_goal():
            yield ServiceActionTerminatedProposition(self.ontology.name, goal.get_device_event())

    def get_goal_attribute(self, goal, attribute):
        plan_info = self._get_plan_info(goal)
        try:
            return plan_info[attribute]
        except KeyError:
            return []

    def get_postplan(self, goal):
        if self.has_goal(goal):
            return self.get_goal_attribute(goal, 'postplan')
        else:
            return []

    def get_superactions(self, goal):
        if self.has_goal(goal):
            return self.get_goal_attribute(goal, 'superactions')
        else:
            return []

    def restart_on_completion(self, goal):
        return self.get_goal_attribute(goal, 'restart_on_completion')

    def get_reraise_on_resume(self, goal):
        reraise_on_resume = self.get_goal_attribute(goal, "reraise_on_resume")
        if reraise_on_resume == []:
            reraise_on_resume = True
        return reraise_on_resume

    def is_silently_accommodatable(self, goal):
        return self.get_goal_attribute(goal, 'unrestricted_accommodation')

    def get_preferred(self, goal):
        if self.has_goal(goal):
            return self.get_goal_attribute(goal, 'preferred')
        else:
            return []

    def goal_allows_accommodation_without_feedback(self, goal):
        return self.get_goal_attribute(goal, "accommodate_without_feedback")

    def get_resolving_answers(self, question):
        def all_answers(question):
            for individual in self.ontology.individuals_as_objects:
                if individual.sort == question.predicate.sort:
                    yield PredicateProposition(question.predicate, individual)

        return unique(all_answers(question))

    def _query_list_to_dict_indexed_by_question(self, query_list):
        plans = {}
        for query_info in query_list:
            query = query_info["query"]
            if query in plans:
                raise InvalidPlansException(f"multiple definitions for query {query}")
            plans[query] = query_info
        return plans

    def _plan_list_to_dict_indexed_by_goal(self, plan_list):
        plans = {}
        for plan_info in plan_list:
            goal = plan_info["goal"]
            if not goal.is_goal():
                raise Exception("expected goal but found %s" % goal)
            if goal in plans:
                raise InvalidPlansException("multiple plans for goal %s" % goal)
            plans[goal] = plan_info
        return plans

    def _add_up_plan(self):
        up_action = self.action("up")
        goal = PerformGoal(up_action)
        self.plans[goal] = {"goal": goal, "plan": Plan()}

    def _add_top_plan_if_missing(self):
        top_goal = PerformGoal(self.action("top"))
        if top_goal not in self.plans:
            self.plans[top_goal] = {"goal": top_goal, "plan": Plan()}

    def get_incremental(self, semantic_object):
        return self._get_parameter(semantic_object, "incremental")

    def get_alternatives(self, semantic_object):
        return self._get_parameter(semantic_object, "alts")

    def get_source(self, semantic_object):
        return self._get_parameter(semantic_object, "source")

    def get_format(self, semantic_object):
        return self._get_parameter(semantic_object, "format")

    def get_service_query(self, semantic_object):
        return self._get_parameter(semantic_object, "service_query")

    def get_on_zero_hits_action(self, semantic_object):
        return self._get_parameter(semantic_object, "on_zero_hits_action")

    def get_on_too_many_hits_action(self, semantic_object):
        return self._get_parameter(semantic_object, "on_too_many_hits_action")

    def get_label_questions(self, semantic_object):
        return self._get_parameter(semantic_object, "label_questions")

    def get_sort_order(self, semantic_object):
        return self._get_parameter(semantic_object, "sort_order")

    def get_max_spoken_alts(self, question):
        return self._get_parameter(question, "max_spoken_alts")

    def get_background(self, semantic_object):
        return self._get_parameter(semantic_object, "background")

    def get_ask_features(self, semantic_object):
        return self._get_parameter(semantic_object, "ask_features")

    def get_hints(self, semantic_object):
        return self._get_parameter(semantic_object, "hints")

    def get_related_information(self, semantic_object):
        return self._get_parameter(semantic_object, "related_information")

    def get_max_reported_hit_count(self, semantic_object):
        return self._get_parameter(semantic_object, "max_reported_hit_count")

    def get_alternatives_predicate(self, goal):
        return self._get_plan_attribute(goal, "alternatives_predicate")

    def get_always_ground(self, semantic_object):
        return self._get_parameter(semantic_object, "always_ground")

    def get_is_always_relevant(self, semantic_object):
        return self._get_parameter(semantic_object, "always_relevant")

    def _get_plan_attribute(self, goal, attribute):
        if goal in self.plans:
            plan_info = self._get_plan_info(goal)
            if attribute in plan_info:
                return plan_info[attribute]

    def get_max_answers(self, goal):
        return self._get_plan_attribute(goal, "max_answers")

    def get_verbalize(self, semantic_object):
        if self._get_parameter(semantic_object, "verbalize") is False:
            return False
        else:
            return True

    def allows_goal_accommodation(self, question):
        if self._get_parameter(question, "allow_goal_accommodation") is False:
            return False
        else:
            return True

    def _get_parameter(self, question, parameter):
        try:
            return self.parameters[question][parameter]
        except KeyError:
            return None

    def get_questions_in_plan(self, plan):
        for item in plan:
            if item.type_ in plan_item.QUESTION_TYPES:
                question = item.content
                for feature_question in self.get_feature_questions_for_plan_item(question, item):
                    yield feature_question
                ask_features = self.get_ask_features(question)
                if ask_features:
                    for ask_feature in ask_features:
                        feature_question = self.ontology.create_wh_question(ask_feature.name)
                        yield feature_question
                        if ask_feature.kpq:
                            yield KnowledgePreconditionQuestion(feature_question)
                yield question

    def get_feature_questions_for_plan_item(self, question, plan_item):
        feature_questions = []
        if question.content.is_lambda_abstracted_predicate_proposition():
            for predicate in list(self.ontology.get_predicates().values()):
                if predicate.is_feature_of(question.content.predicate):
                    feature_question = self.ontology.create_wh_question(predicate.get_name())
                    feature_questions.append(feature_question)
        return feature_questions

    def is_question_in_plan(self, question, plan):
        return question in self.get_questions_in_plan(plan)

    def get_invoke_service_action_items_for_action(self, action_name):
        for goal in self.get_all_goals():
            plan = self.get_plan(goal)
            for item in plan:
                if item.type_ == plan_item.TYPE_INVOKE_SERVICE_ACTION and item.get_service_action() == action_name:
                    yield item

    def get_names_of_user_targeted_actions(self):
        actions = []
        for goal in self.get_all_goals():
            plan = self.get_plan(goal)
            for item in plan:
                if item.type_ == plan_item.TYPE_GET_DONE:
                    actions.append(item.content.value)
        return actions

    def get_implications_for_domain_query(self, query):
        return self.queries[query]["implications"]

    def get_propositions_for_random_selection(self, query):
        return self.queries[query]["for_random_selection"]

    def get_propositions_for_enumeration(self, query):
        return copy.copy(self.queries[query].get("for_enumeration", []))

    def get_propositions_for_random_enumeration(self, query):
        return copy.copy(self.queries[query].get("for_random_enumeration", []))

    def get_limit_for_iterator(self, query):
        limit = self.queries[query].get("limit", -1)
        return int(limit)



import warnings
import re

from tala.model.semantic_object import SemanticObject, OntologySpecificSemanticObject
from tala.utils.as_semantic_expression import AsSemanticExpressionMixin
from tala.model.polarity import Polarity


class Individual(OntologySpecificSemanticObject, AsSemanticExpressionMixin):
    def __init__(self, ontology_name, value, sort):
        OntologySpecificSemanticObject.__init__(self, ontology_name)
        if sort.is_string_sort():
            value = self._strip_quotes(value)
        self.value = value
        self.sort = sort
        self.polarity = Polarity.POS

    def getValue(self):
        warnings.warn(
            "Individual.getValue() is deprecated. Use Individual.value instead.", DeprecationWarning, stacklevel=2
        )
        return self.value

    def getSort(self):
        warnings.warn(
            "Individual.getSort() is deprecated. Use Individual.sort instead.", DeprecationWarning, stacklevel=2
        )
        return self.sort

    def is_individual(self):
        return True

    def is_positive(self):
        return True

    def __eq__(self, other):
        try:
            if other.is_positive():
                return self.value == other.value and self.sort == other.sort
            else:
                return False
        except AttributeError:
            return False

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash((self.value, self.sort))

    def __str__(self):
        sort = self.sort
        if sort.is_string_sort():
            return '"%s"' % self.value
        else:
            return str(self.value)

    def __repr__(self):
        return "%s%s" % (self.__class__.__name__, (self.value, self.sort))

    def negate(self):
        return NegativeIndividual(self.ontology_name, self.value, self.sort)

    def _strip_quotes(self, string):
        m = re.search('^"([^"]*)"$', string)
        if m:
            string_content = m.group(1)
            return string_content
        else:
            return string

    def value_as_json_object(self):
        return self.sort.value_as_json_object(self.value)


class NegativeIndividual(Individual):
    def __init__(self, ontology_name, value, sort):
        OntologySpecificSemanticObject.__init__(self, ontology_name)
        if sort.is_string_sort():
            value = self._strip_quotes(value)
        self.value = value
        self.sort = sort
        self.polarity = Polarity.NEG

    def negate(self):
        return Individual(self.ontology_name, self.value, self.sort)

    def __str__(self):
        return "~%s" % self.value

    def __eq__(self, other):
        try:
            if other.is_positive():
                return False
            else:
                otherValue = other.value
                return self.value == otherValue
        except AttributeError:
            return False

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return super(NegativeIndividual, self).__hash__()

    def is_positive(self):
        return False


class Yes(SemanticObject, AsSemanticExpressionMixin):
    YES = "yes"

    def is_positive(self):
        return True

    def is_yes(self):
        return True

    def __str__(self):
        return Yes.YES

    def __eq__(self, other):
        try:
            return self.YES == other.YES
        except:  # noqa: E722
            return False

    def __ne__(self, other):
        try:
            return self.YES != other.YES
        except:  # noqa: E722
            return False

    def __hash__(self):
        return hash(str(self))

    def as_dict(self):
        result = {"semantic_object_type": "yes/no", "instance": self.YES}
        return super().as_dict() | result


class No(SemanticObject, AsSemanticExpressionMixin):
    NO = "no"

    def is_positive(self):
        return False

    def is_no(self):
        return True

    def __str__(self):
        return No.NO

    def __eq__(self, other):
        try:
            return self.NO == other.NO
        except:  # noqa: E722
            return False

    def __ne__(self, other):
        try:
            return self.NO != other.NO
        except:  # noqa: E722
            return False

    def __hash__(self):
        return hash(str(self))

    def as_dict(self):
        result = {"semantic_object_type": "yes/no", "instance": self.NO}
        return super().as_dict() | result



class AbstractServiceWrapper(object):
    def __init__(self):
        super(AbstractServiceWrapper, self).__init__()
        self._session_data = {}

    def query(self, question, parameters, min_results, max_results, session, context):
        raise NotImplementedError()

    def perform(self, action_name, parameters, session, context):
        raise NotImplementedError()

    def validate(self, validator_name, parameters, session, context):
        raise NotImplementedError()

    @property
    def session_data(self):
        return self._session_data

    @session_data.setter
    def session_data(self, new_session_data):
        self._session_data = new_session_data



import unittest

import pytest

from tala.ddd.services.service_interface import ServiceActionInterface, ServiceInterface, \
    UnexpectedActionException, ServiceQueryInterface, ServiceValidatorInterface, \
    DuplicateNameException, PlayAudioActionInterface, AudioURLServiceParameter, FrontendTarget, HttpTarget, \
    FailureReasonsNotAllowedException, ActionFailureReason, UnsupportedServiceInterfaceTarget
from tala.testing.utils import EqualityAssertionTestCaseMixin

from tala.ddd.json_parser import JSONServiceInterfaceParser

if 'unittest.util' in __import__('sys').modules:
    # Show full diff in self.assertEqual.
    __import__('sys').modules['unittest.util']._MAX_LENGTH = 999999999


class ServiceInterfaceTests(unittest.TestCase):
    def setUp(self):
        self._play_audio_actions = []
        self._custom_actions = []
        self._queries = []
        self._validators = []
        self._http_target = HttpTarget("mock_endpoint")
        self._audio_url_parameter = AudioURLServiceParameter("audio_url")
        self._service_interface = None
        self._result = None

    def test_has_play_audio_action_when_action_exists(self):
        self._given_play_audio_actions(["mocked_action"])
        self._given_service_interface()
        self._when_calling_has_action("mocked_action")
        self._then_result_is(True)

    def _create_play_audio_action(self, name):
        return PlayAudioActionInterface(
            name, FrontendTarget(), parameters=[], audio_url_parameter=self._audio_url_parameter
        )

    def _given_play_audio_actions(self, names):
        self._play_audio_actions = [self._create_play_audio_action(name) for name in names]

    def test_has_action_when_action_exists(self):
        self._given_actions(["mocked_action"])
        self._given_service_interface()
        self._when_calling_has_action("mocked_action")
        self._then_result_is(True)

    def _create_action(self, name):
        return ServiceActionInterface(name, self._http_target, parameters=[], failure_reasons=[])

    def _given_actions(self, names):
        self._custom_actions = [self._create_action(name) for name in names]

    def _given_service_interface(self):
        self._create_service_interface()

    def _create_service_interface(self):
        actions = self._play_audio_actions + self._custom_actions
        service_interface = ServiceInterface(actions, self._queries, self._validators)
        self._service_interface = JSONServiceInterfaceParser().parse(service_interface.as_json())

    def _when_calling_has_action(self, name):
        self._result = self._service_interface.has_action(name)

    def _then_result_is(self, expected_result):
        self.assertEqual(self._result, expected_result)

    def test_has_play_audio_action_when_action_missing(self):
        self._given_play_audio_actions(["mocked_action"])
        self._given_service_interface()
        self._when_calling_has_action("another_action")
        self._then_result_is(False)

    def test_has_action_when_action_missing(self):
        self._given_actions(["mocked_action"])
        self._given_service_interface()
        self._when_calling_has_action("another_action")
        self._then_result_is(False)

    def test_has_play_audio_action_when_action_exists_among_several(self):
        self._given_play_audio_actions(["mocked_action", "another_action", "third_action"])
        self._given_service_interface()
        self._when_calling_has_action("another_action")
        self._then_result_is(True)

    def test_has_action_when_action_exists_among_several(self):
        self._given_actions(["mocked_action", "another_action", "third_action"])
        self._given_service_interface()
        self._when_calling_has_action("another_action")
        self._then_result_is(True)

    def test_get_play_audio_action_when_action_exists(self):
        self._given_play_audio_actions(["mocked_action"])
        self._given_service_interface()
        self._when_calling_get_action("mocked_action")
        self._then_result_is_play_audio_action("mocked_action")

    def _then_result_is_play_audio_action(self, name):
        expected_action = self._create_play_audio_action(name)
        self.assertEqual(self._result, expected_action)

    def test_get_action_when_action_exists(self):
        self._given_actions(["mocked_action"])
        self._given_service_interface()
        self._when_calling_get_action("mocked_action")
        self._then_result_is_action_named("mocked_action")

    def _when_calling_get_action(self, name):
        self._result = self._service_interface.get_action(name)

    def _then_result_is_action_named(self, name):
        expected_action = self._create_action(name)
        self.assertEqual(self._result, expected_action)

    def test_get_play_audio_action_when_action_missing(self):
        self._given_play_audio_actions(["mocked_action"])
        self._given_service_interface()
        self._when_calling_get_action_then_exception_is_raised_matching(
            "another_action", UnexpectedActionException,
            r"Expected one of the known actions \['mocked_action'\] but got 'another_action'"
        )

    def test_get_action_when_action_missing(self):
        self._given_actions(["mocked_action"])
        self._given_service_interface()
        self._when_calling_get_action_then_exception_is_raised_matching(
            "another_action", UnexpectedActionException,
            r"Expected one of the known actions \['mocked_action'\] but got 'another_action'"
        )

    def _when_calling_get_action_then_exception_is_raised_matching(self, name, expected_exception, expected_message):
        with self.assertRaisesRegex(expected_exception, expected_message):
            self._service_interface.get_action(name)

    def test_duplicate_actions(self):
        self._given_actions(["mocked_action", "mocked_action"])
        self._when_creating_service_interface_then_exception_is_raised_matching(
            DuplicateNameException, "Expected all names to be unique among "
            r"\[ServiceActionInterface\('mocked_action', HttpTarget\('mock_endpoint'\), parameters=\[\], failure_reasons=\[\]\),"
            r" ServiceActionInterface\('mocked_action', HttpTarget\('mock_endpoint'\), parameters=\[\], failure_reasons=\[\]\)\] "
            "but they weren't"
        )

    def _when_creating_service_interface_then_exception_is_raised_matching(self, expected_exception, expected_message):
        with self.assertRaisesRegex(expected_exception, expected_message):
            self._given_service_interface()

    def test_duplicate_queries(self):
        self._given_queries(["mocked_query", "another_query", "mocked_query"])
        self._when_creating_service_interface_then_exception_is_raised_matching(
            DuplicateNameException, "Expected all names to be unique among .* but they weren't"
        )

    def _given_queries(self, names):
        self._queries = [self._create_query(name) for name in names]

    def _create_query(self, name):
        return ServiceQueryInterface(name, self._http_target, [])

    def test_duplicate_validators(self):
        self._given_validators(["validator", "validator", "other"])
        self._when_creating_service_interface_then_exception_is_raised_matching(
            DuplicateNameException, "Expected all names to be unique among .* but they weren't"
        )

    def _given_validators(self, names):
        self._validators = [self._create_validator(name) for name in names]

    def _create_validator(self, name):
        return ServiceValidatorInterface(name, self._http_target, [])


class ActionInterfaceTests(unittest.TestCase):
    def setUp(self):
        self._action = None
        self._result = None

    def test_action_interface(self):
        self._given_action(ServiceActionInterface("mocked_action", FrontendTarget(), [], []))
        self._when_asking_is_play_audio_action()
        self._then_result_is(False)

    def _given_action(self, action):
        self._action = action

    def _then_result_is(self, expected_result):
        self.assertEqual(self._result, expected_result)

    def _when_asking_is_play_audio_action(self):
        self._result = self._action.is_play_audio_action

    def test_play_audio_action_interface(self):
        self._given_action(PlayAudioActionInterface("mocked_action", FrontendTarget(), [], "mocked_url"))
        self._when_asking_is_play_audio_action()
        self._then_result_is(True)


class ActionInterfaceCreationTests(unittest.TestCase):
    def setUp(self):
        self._name = None
        self._target = None
        self._parameters = []
        self._failure_reasons = []
        self._result = None

    def test_failure_reasons_not_allowed_with_frontend_target(self):
        self._given_name("MockedAction")
        self._given_target(FrontendTarget())
        self._given_failure_reasons([ActionFailureReason("mocked_failure_reason")])
        self._when_creating_action_then_exception_is_raised_matching(
            FailureReasonsNotAllowedException, "Expected no failure reasons for action 'MockedAction' with target "
            r"'frontend', but got \['mocked_failure_reason'\]"
        )

    def _given_name(self, name):
        self._name = name

    def _given_target(self, target):
        self._target = target

    def _given_failure_reasons(self, failure_reasons):
        self._failure_reasons = failure_reasons

    def _when_creating_action_then_exception_is_raised_matching(self, expected_exception, expected_pattern):
        with self.assertRaisesRegex(expected_exception, expected_pattern):
            self._create_action()

    def test_failure_reasons_allowed_with_http_target(self):
        self._given_name("MockedAction")
        self._given_target(HttpTarget("mock_endpoint"))
        self._given_failure_reasons([ActionFailureReason("mocked_failure_reason")])
        self._when_creating_action()
        self._then_result_matches(
            ServiceActionInterface(
                "MockedAction", HttpTarget("mock_endpoint"), [], [ActionFailureReason("mocked_failure_reason")]
            )
        )

    def _when_creating_action(self):
        self._result = self._create_action()

    def _create_action(self):
        return ServiceActionInterface(self._name, self._target, self._parameters, self._failure_reasons)

    def _then_result_matches(self, expected_result):
        self.assertEqual(self._result, expected_result)


class TestExceptionRaisedForUnsupportedTarget(object):
    def create_action_interface(self, target):
        ServiceActionInterface("mock_name", target, "mock_parameters", failure_reasons=[])

    def create_validator_interface(self, target):
        ServiceValidatorInterface("mock_name", target, "mock_parameters")

    def create_query_interface(self, target):
        ServiceQueryInterface("mock_name", target, "mock_parameters")

    http_target = HttpTarget("mock_endpoint")
    frontend_target = FrontendTarget()

    @pytest.mark.parametrize(
        "interface_constructor,target", [
            (create_action_interface, http_target),
            (create_action_interface, frontend_target),
            (create_query_interface, http_target),
            (create_validator_interface, http_target),
        ]
    )
    def test_supported_targets(self, interface_constructor, target):
        self.when_create_interface(interface_constructor, target)
        self.then_no_exception_is_raised()

    def when_create_interface(self, interface_constructor, target):
        interface_constructor(self, target)

    def then_no_exception_is_raised(self):
        pass

    @pytest.mark.parametrize("interface_constructor", [create_query_interface, create_validator_interface])
    def test_unsupported_frontend_target(self, interface_constructor):
        self.when_create_interface_then_exception_is_raised_matching(
            interface_constructor, FrontendTarget(), UnsupportedServiceInterfaceTarget,
            "Expected a non-frontend target for service interface 'mock_name' but got a frontend target."
        )

    def when_create_interface_then_exception_is_raised_matching(
        self, interface_constructor, target, expected_exception, expected_message
    ):
        with pytest.raises(expected_exception) as excinfo:
            interface_constructor(self, target)
        assert expected_message == str(excinfo.value)


class ServiceTargetsTestCase(unittest.TestCase, EqualityAssertionTestCaseMixin):
    def test_frontend_target_equality(self):
        self.assert_eq_returns_true_and_ne_returns_false_symmetrically(FrontendTarget(), FrontendTarget())

    def test_frontend_target_target_not_equals_instance_of_other_class(self):
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(FrontendTarget(), "instance_of_other_class")

    def test_http_target_equality(self):
        self.assert_eq_returns_true_and_ne_returns_false_symmetrically(
            HttpTarget("mock_endpoint"), HttpTarget("mock_endpoint")
        )

    def test_http_target_inequality_due_to_endpoint(self):
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(
            HttpTarget("mock_endpoint_1"), HttpTarget("mock_endpoint_2")
        )

    def test_http_target_target_not_equals_instance_of_other_class(self):
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(
            HttpTarget("mock_endpoint"), "instance_of_other_class"
        )



class ParameterBindingsFormatter:
    @staticmethod
    def to_values(bindings):
        return [binding.value for binding in bindings]

    @staticmethod
    def to_proposition_list(bindings):
        def get_propositions():
            for binding in bindings:
                for proposition in binding.propositions:
                    if proposition:
                        yield proposition

        return list(get_propositions())



from unittest.runner import TextTestResult

DESCRIPTIONS = {
    "FAIL": "Failure in test '{}'",
    "ERROR": "Error in test '{}'",
}


class InteractionTestResult(TextTestResult):
    def addFailure(self, test, err):
        """Called when an error has occurred. 'err' is a tuple of values as
        returned by sys.exc_info()."""
        type_, exception, traceback = err
        self.failures.append((test, str(exception)))
        self._mirrorOutput = True
        if self.showAll:
            self.stream.writeln("FAIL")
        elif self.dots:
            self.stream.write('F')
            self.stream.flush()

    def printErrorList(self, flavour, errors):
        for test, err in errors:
            self.stream.writeln(self.separator1)
            self.stream.writeln(DESCRIPTIONS[flavour].format(test.name))
            self.stream.writeln(self.separator2)
            self.stream.writeln(str(err))
            self.stream.writeln()



# -*- coding: utf-8 -*-

import json
import pytest
import unittest
from io import StringIO

from mock import patch, ANY, call

from tala.testing.interactions import compiler
from tala.testing.interactions.compiler import InteractionTestCompiler, ParseException

TESTS_FILENAME = "tdm/test/test_interactiontest.txt"


class TestFileParsing:
    def test_is_turn(self):
        assert InteractionTestCompiler._is_turn("U> ")
        assert InteractionTestCompiler._is_turn("U> ''")
        assert InteractionTestCompiler._is_turn("S> ['ask(?X.goal(X)))']")
        assert InteractionTestCompiler._is_turn("U> ['answer(yes)']")
        assert InteractionTestCompiler._is_turn("U> hello there $CHECK")
        assert InteractionTestCompiler._is_turn("S> hello there")
        assert not InteractionTestCompiler._is_turn("")
        assert not InteractionTestCompiler._is_turn("sdflkjsdf")


class TestCompiler():
    def setup_method(self):
        self._compiler = InteractionTestCompiler()

    @patch("%s.InteractionTest" % compiler.__name__)
    def test_name(self, InteractionTest):
        self.when_compile("""\
--- first test
S> system utterance

--- second test
U> user input
""")
        self.then_interaction_tests_created_with(InteractionTest, ["first test", "second test"])

    def then_interaction_tests_created_with(self, InteractionTest, expected_names):
        expected_calls = [call(ANY, name, ANY) for name in expected_names]
        InteractionTest.assert_has_calls(expected_calls, any_order=True)

    def when_compile(self, string):
        file_ = StringIO(string)
        self._result = self._compiler.compile_interaction_tests("mock_test_file.txt", file_)

    def test_user_passivity(self):
        self.when_compile("""\
--- mock name
U>
""")
        self.then_result_has_user_passivity_turn_with_properties(2)

    def then_result_has_user_passivity_turn_with_properties(self, expected_line_number):
        actual_turn = self._get_actual_single_turn_from_single_test()
        assert actual_turn.is_user_passivity_turn
        assert expected_line_number == actual_turn.line_number

    def _get_actual_single_turn_from_single_test(self):
        assert 1 == len(self._result)
        actual_test = self._result[0]
        assert 1 == len(actual_test.turns)
        return actual_test.turns[0]

    def test_user_moves(self):
        self.when_compile("""\
--- mock name
U> ["request(make_reservation)", "answer(paris)"]
""")
        self.then_result_is_user_interpretation()

    def test_moves_of_user_moves(self):
        self.when_compile("""\
--- mock name
U> ["request(make_reservation)", "answer(paris)"]
""")
        self.then_result_has_moves(["request(make_reservation)", "answer(paris)"])

    def test_rich_user_moves(self):
        self.when_compile(
            """\
--- mock name
U> [{"semantic_expression": "request(make_reservation)", "understanding_confidence": 0.12, "perception_confidence": 0.23}, {"semantic_expression": "answer(paris)", "understanding_confidence": 0.45, "perception_confidence": 0.67}]
"""
        )
        self.then_result_has_moves([{
            "semantic_expression": "request(make_reservation)",
            "understanding_confidence": 0.12,
            "perception_confidence": 0.23
        }, {
            "semantic_expression": "answer(paris)",
            "understanding_confidence": 0.45,
            "perception_confidence": 0.67
        }])

    def test_line_number_of_user_moves(self):
        self.when_compile("""\
--- mock name
U> ["request(make_reservation)", "answer(paris)"]
""")
        self.then_result_has_line_number(2)

    def test_utterance_of_user_moves(self):
        self.when_compile("""\
--- mock name
U> ["request(make_reservation)", "answer(paris)"]
""")

        self.then_result_has_utterance(None)

    def test_modality_of_user_moves(self):
        self.when_compile("""\
--- mock name
U> ["request(make_reservation)", "answer(paris)"]
""")

        self.then_result_has_modality(None)

    def test_user_interpretation(self):
        self.when_compile(
            """\
--- mock name
U> {"moves": ["request(make_reservation)", "answer(paris)"], "utterance": "an utterance", "modality": "speech"}
"""
        )
        self.then_result_is_user_interpretation()

    def test_user_interpretations(self):
        self.when_compile(
            """\
--- mock name
U> {"interpretations": [{"moves": ["request(make_reservation)", "answer(paris)"], "utterance": "an utterance", "modality": "speech"}], "entitites": []}
"""
        )
        self.then_result_is_semantic_input()

    def then_result_is_semantic_input(self):
        actual_turn = self._get_actual_single_turn_from_single_test()
        assert actual_turn.is_user_semantic_input_turn

    @patch(f"{compiler.__name__}.json", autospec=True)
    def test_invalid_json(self, mock_json):
        self.given_json_loads_raises_decode_error(mock_json, "the error", doc="doc", pos=0)
        self.when_compile_then_exception_is_raised(
            string="""\
--- mock name
U> {"moves": []}
""",
            expected_exception_class=ParseException,
            expected_message=r"Expected valid JSON on line 2 of 'mock_test_file.txt' but encountered a decoding error."
            r"\n\n  Line 2: {\"moves\": \[\]}\n\n  Error: 'the error: line 1 column 1 \(char 0\)'"
        )

    def given_json_loads_raises_decode_error(self, mock_json, *args, **kwargs):
        mock_json.decoder.JSONDecodeError = json.decoder.JSONDecodeError
        mock_json.loads.side_effect = json.decoder.JSONDecodeError(*args, **kwargs)

    def when_compile_then_exception_is_raised(self, string, expected_exception_class, expected_message):
        with pytest.raises(expected_exception_class, match=expected_message):
            self.when_compile(string)

    def test_moves_of_user_interpretation(self):
        self.when_compile(
            """\
--- mock name
U> {"moves": ["request(make_reservation)", "answer(paris)"], "utterance": "an utterance", "modality": "speech"}
"""
        )
        self.then_result_has_moves(["request(make_reservation)", "answer(paris)"])

    def test_line_number_of_user_interpretation(self):
        self.when_compile(
            """\
--- mock name
U> {"moves": ["request(make_reservation)", "answer(paris)"], "utterance": "an utterance", "modality": "speech"}
"""
        )
        self.then_result_has_line_number(2)

    def test_utterance_of_user_interpretation(self):
        self.when_compile(
            """\
--- mock name
U> {"moves": ["request(make_reservation)", "answer(paris)"], "utterance": "an utterance", "modality": "speech"}
"""
        )

        self.then_result_has_utterance("an utterance")

    def test_modality_of_user_interpretation(self):
        self.when_compile(
            """\
--- mock name
U> {"moves": ["request(make_reservation)", "answer(paris)"], "utterance": "an utterance", "modality": "speech"}
"""
        )

        self.then_result_has_modality("speech")

    def then_result_is_user_interpretation(self):
        actual_turn = self._get_actual_single_turn_from_single_test()
        assert actual_turn.is_user_interpretation_turn

    def then_result_has_moves(self, expected_moves):
        actual_turn = self._get_actual_single_turn_from_single_test()
        assert expected_moves == actual_turn.moves

    def then_result_has_line_number(self, expected_line_number):
        actual_turn = self._get_actual_single_turn_from_single_test()
        assert expected_line_number == actual_turn.line_number

    def then_result_has_utterance(self, expected):
        actual_turn = self._get_actual_single_turn_from_single_test()
        assert expected == actual_turn.utterance

    def then_result_has_modality(self, expected):
        actual_turn = self._get_actual_single_turn_from_single_test()
        assert expected == actual_turn.modality

    def test_system_moves(self):
        self.when_compile("""\
--- mock name
S> ["request(make_reservation)", "answer(paris)"]
""")
        self.then_result_is_system_moves()

    def test_moves_of_system_moves(self):
        self.when_compile("""\
--- mock name
S> ["request(make_reservation)", "answer(paris)"]
""")
        self.then_result_has_moves(["request(make_reservation)", "answer(paris)"])

    def test_properties_of_system_moves(self):
        self.when_compile("""\
--- mock name
S> ["request(make_reservation)", "answer(paris)"]
""")
        self.then_result_has_line_number(expected_line_number=2)

    def then_result_is_system_moves(self):
        actual_turn = self._get_actual_single_turn_from_single_test()
        assert actual_turn.is_system_moves_turn

    def test_multiple_turns(self):
        self.when_compile(
            """\
--- mock name
S> What do you want to do?
U> ["request(make_reservation)", "answer(paris)"]
"""
        )
        self.then_turns_are_system_utterance_and_user_moves()

    def then_turns_are_system_utterance_and_user_moves(self):
        actual_test = self._result[0]
        assert actual_test.turns[0].is_system_output_turn
        assert actual_test.turns[1].is_user_interpretation_turn

    def test_system_utterance(self):
        self.when_compile("""\
--- mock name
S> Welcome.
""")
        self.then_result_has_system_utterance_turn_with_properties("Welcome.", 2)

    def then_result_has_system_utterance_turn_with_properties(self, expected_utterance, expected_line_number):
        actual_turn = self._get_actual_single_turn_from_single_test()
        assert actual_turn.is_system_utterance_turn
        assert expected_utterance == actual_turn.utterance
        assert expected_line_number == actual_turn.line_number

    def test_single_recognition_hypothis(self):
        self.when_compile("""\
--- mock name
U> to paris
""")
        self.then_result_has_recognition_hypotheses_turn_with_properties([("to paris", None)], 2)

    @pytest.mark.parametrize("utterance", ["''", '""'])
    def test_empty_user_utterance(self, utterance):
        self.when_compile(f"""\
--- mock name
U> {utterance}
""")
        self.then_result_has_recognition_hypotheses_turn_with_properties([("", None)], 2)

    def then_result_has_recognition_hypotheses_turn_with_properties(self, expected_hypotheses, expected_line_number):
        actual_turn = self._get_actual_single_turn_from_single_test()
        assert actual_turn.is_recognition_hypotheses_turn
        assert expected_hypotheses == actual_turn.hypotheses
        assert expected_line_number == actual_turn.line_number

    def test_recognition_hypothesis_with_confidence_level(self):
        self.when_compile("""\
--- mock name
U> to paris $CHECK
""")
        self.then_result_has_recognition_hypotheses_turn_with_properties([("to paris", "$CHECK")], 2)

    def test_multiple_recognition_hypotheses(self):
        self.when_compile("""\
--- mock name
U> first hypothesis 0.6 | second hypothesis 0.5
""")
        self.then_result_has_recognition_hypotheses_turn_with_properties([("first hypothesis", "0.6"),
                                                                          ("second hypothesis", "0.5")], 2)

    def test_notify_started(self):
        self.when_compile(
            """\
--- mock name
Event> {"name": "AlarmRings", "status": "started", "parameters": {"alarm_hour": 7, "alarm_minute": 30}}
"""
        )
        self.then_result_has_notify_started_turn_with_properties("AlarmRings", {"alarm_hour": 7, "alarm_minute": 30}, 2)

    def then_result_has_notify_started_turn_with_properties(
        self, expected_action, expected_parameters, expected_line_number
    ):
        actual_turn = self._get_actual_single_turn_from_single_test()
        assert actual_turn.is_notify_started_turn
        assert expected_action == actual_turn.action
        assert expected_parameters == actual_turn.parameters
        assert expected_line_number == actual_turn.line_number

    def test_unicode_strings(self):
        self.when_compile("""\
--- mock name
S> Vad vill du göra?
""")
        self.then_result_has_system_utterance_turn_with_properties("Vad vill du göra?", 2)

    def test_empty_system_utterance(self):
        self.when_compile("""\
--- mock name
S>
""")
        self.then_result_has_system_utterance_turn_with_properties("", 2)


class PrettyFormattingTests(unittest.TestCase):
    def setUp(self):
        self._hypotheses = []
        self._system_utterance = None

    def test_pretty_hypotheses(self):
        self._given_hypothesis("first utterance", 0.9)
        self._given_hypothesis("second utterance", 0.8)
        self._when_formatting_hypotheses()
        self._then_the_result_is("U> first utterance 0.9 | second utterance 0.8")

    def _given_hypothesis(self, utterance, confidence):
        self._hypotheses.append((utterance, confidence))

    def _when_formatting_hypotheses(self):
        self._result = InteractionTestCompiler.pretty_hypotheses(self._hypotheses)

    def _then_the_result_is(self, expected_result):
        assert self._result == expected_result

    def test_pretty_passivity(self):
        self._when_formatting_passivity()
        self._then_the_result_is("U>")

    def _when_formatting_passivity(self):
        self._result = InteractionTestCompiler.pretty_passivity()

    def test_pretty_system_utterance(self):
        self._given_system_utterance("an utterance")
        self._when_formatting_system_utterance()
        self._then_the_result_is("S> an utterance")

    def _given_system_utterance(self, utterance):
        self._system_utterance = utterance

    def _when_formatting_system_utterance(self):
        self._result = InteractionTestCompiler.pretty_system_utterance(self._system_utterance)
